import re
from datetime import datetime

import dateutil.parser
from numpy import nan
from packaging import version


class NONMEMResultsFile:
    """Representing and parsing a NONMEM results file (aka lst-file)
    This is not a generic output object and will be combined by other classes
    into new structures.
    We rely on tags in NONMEM >= 7.2.0
    """

    def __init__(self, path=None, log=None):
        self.log = log
        self.table = dict()
        self.nonmem_version = None
        self.runtime_total = None
        if path is not None:
            for name, content in self.table_blocks(path):
                if name == 'INIT':
                    self.nonmem_version = content.pop('nonmem_version', None)
                elif name == 'runtime':
                    self.runtime_total = content.pop('total', None)
                else:
                    self.table[name] = content

    @property
    def _supported_nonmem_version(self):
        return NONMEMResultsFile.supported_version(self.nonmem_version)

    def estimation_status(self, table_number):
        result = NONMEMResultsFile.unknown_termination()
        if self._supported_nonmem_version:
            if table_number in self.table.keys():
                for key in result.keys():
                    result[key] = self.table[table_number].get(key)
            else:
                result['minimization_successful'] = False
        return result

    def covariance_status(self, table_number):
        result = NONMEMResultsFile.unknown_covariance()
        if self._supported_nonmem_version:
            if table_number in self.table.keys():
                for key in result.keys():
                    result[key] = self.table[table_number].get(key)
            else:
                result['covariance_step_ok'] = False
        return result

    def ofv(self, table_number):
        ofv = None
        if self._supported_nonmem_version:
            if table_number in self.table.keys():
                ofv = self.table[table_number].get('OBJV')
        if ofv is not None:
            ofv = float(ofv)
        else:
            ofv = nan
        return ofv

    @staticmethod
    def supported_version(nonmem_version):
        return nonmem_version is not None and version.parse(nonmem_version) >= version.parse(
            '7.2.0'
        )

    @staticmethod
    def unknown_covariance():
        return {'covariance_step_ok': None}

    @staticmethod
    def unknown_termination():
        return {
            'minimization_successful': None,
            'estimate_near_boundary': None,
            'rounding_errors': None,
            'maxevals_exceeded': None,
            'significant_digits': nan,
            'function_evaluations': nan,
            'warning': None,
        }

    @staticmethod
    def cleanup_version(v):
        if v == 'V':
            v = '5.0'
        elif v == 'VI':
            v = '6.0'
        return v

    @staticmethod
    def parse_covariance(rows):
        result = NONMEMResultsFile.unknown_covariance()
        result['covariance_step_ok'] = False
        if len(rows) < 1:
            return result

        not_ok = re.compile(r' INTERPRET VARIANCE-COVARIANCE OF ESTIMATES WITH CARE')
        not_ok_either = re.compile(r'(R|S) MATRIX ALGORITHMICALLY SINGULAR')
        ok = re.compile(r' Elapsed covariance\s+time in seconds: ')  # need variable whitespace

        for row in rows:
            if not_ok.match(row):
                result['covariance_step_ok'] = False
                break
            if not_ok_either.search(row):
                result['covariance_step_ok'] = False
                break
            if ok.match(row):
                result['covariance_step_ok'] = True
                break
        return result

    @staticmethod
    def parse_termination(rows):
        result = NONMEMResultsFile.unknown_termination()
        if len(rows) < 1:  # will happen if e.g. TERMINATED BY OBJ during estimation
            result['minimization_successful'] = False
            return result
        result['estimate_near_boundary'] = False
        result['rounding_errors'] = False
        result['maxevals_exceeded'] = False
        result['warning'] = False

        success = [
            re.compile(r'0MINIMIZATION SUCCESSFUL'),
            re.compile(
                r' (REDUCED STOCHASTIC PORTION|OPTIMIZATION|'
                + r'BURN-IN|EXPECTATION ONLY PROCESS)'
                + r'( STATISTICALLY | WAS | )(COMPLETED|NOT TESTED)'
            ),
            re.compile(r'1OBJECTIVE FUNCTION IS TO BE EVALUATED'),
        ]
        failure = [
            re.compile(r'0MINIMIZATION TERMINATED'),
            re.compile(r'0SEARCH WITH ESTIMATION STEP WILL NOT PROCEED'),
            re.compile(r'\s*INDIVIDUAL OBJECTIVE FUNCTION VALUES ARE ALL ZERO\. PROBLEM ENDED'),
        ]
        maybe = re.compile(
            r' (REDUCED STOCHASTIC PORTION|OPTIMIZATION|'
            + r'BURN-IN|EXPECTATION ONLY PROCESS)'
            + r'( WAS | )NOT COMPLETED'
        )  # success only if next line USER INTERRUPT
        misc = {
            'estimate_near_boundary': re.compile(
                r'0(ESTIMATE OF THETA IS NEAR THE BOUNDARY AND|'
                + r'PARAMETER ESTIMATE IS NEAR ITS BOUNDARY)'
            ),
            'rounding_errors': re.compile(r' DUE TO ROUNDING ERRORS'),
            'maxevals_exceeded': re.compile(r' DUE TO MAX. NO. OF FUNCTION EVALUATIONS EXCEEDED'),
            'warning': re.compile(r' HOWEVER, PROBLEMS OCCURRED WITH THE MINIMIZATION.'),
        }
        sig_digits = re.compile(r' NO. OF SIG. DIGITS IN FINAL EST.:\s*(\S+)')  # only classical est
        feval = re.compile(r' NO. OF FUNCTION EVALUATIONS USED:\s*(\S+)')  # only classical est

        maybe_success = False
        for row in rows:
            if maybe_success:
                result['minimization_successful'] = bool(re.search(r'USER INTERRUPT', row))
                break
            for p in success:
                if p.match(row):
                    result['minimization_successful'] = True
                    break
            if result['minimization_successful'] is not None:
                break
            for p in failure:
                if p.match(row):
                    result['minimization_successful'] = False
                    break
            if result['minimization_successful'] is not None:
                break
            maybe_success = bool(maybe.match(row))
        for row in rows:
            if sig_digits.match(row):
                result['significant_digits'] = float(sig_digits.match(row).group(1))
                continue
            if feval.match(row):
                result['function_evaluations'] = int(feval.match(row).group(1))
                continue
            for name, p in misc.items():
                if p.match(row):
                    result[name] = True
                    break
        return result

    @staticmethod
    def parse_runtime(row, row_next=None):
        # TODO: support AM/PM
        weekday_month_en = re.compile(
            r'^\s*(Sun|Mon|Tue|Wed|Thu|Fri|Sat)'
            r'\s+(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)'  # Month
            r'\s+(\d+)'  # Day
            r'\s+.*'
            r'\s+(\d{4})'  # Year
        )
        weekday_month_sv = re.compile(
            r'^\s*(mån|tis|ons|tor|fre|lör|sön)'
            r'\s+(\d+)'  # Day
            r'\s+(jan|feb|mar|apr|maj|jun|jul|aug|sep|okt|nov|dec)'  # Month
            r'\s+(\d+)'  # Year
        )
        day_month_year = re.compile(r'^(\d{2})/(\d{2})/(\d{4})\s*$')  # dd/mm/yyyy
        year_month_day = re.compile(r'^(\d{4})-\d{2}-\d{2}\s*$')  # yyyy/mm/dd
        timestamp = re.compile(r'([0-9]{2}:[0-9]{2}:[0-9]{2})')

        month_no = {
            'JAN': 1,
            'FEB': 2,
            'MAR': 3,
            'APR': 4,
            'MAY': 5,
            'JUN': 6,
            'JUL': 7,
            'AUG': 8,
            'SEP': 9,
            'OCT': 10,
            'NOV': 11,
            'DEC': 12,
        }
        month_trans = {'MAJ': 'MAY', 'OKT': 'OCT'}

        date_time = None
        if weekday_month_en.match(row) or weekday_month_sv.match(row):
            if weekday_month_en.match(row):
                _, month, day, year = weekday_month_en.match(row).groups()
            else:
                _, day, month, year = weekday_month_sv.match(row).groups()

            try:
                month = month_no[month.upper()]
            except KeyError:
                month_en = month_trans[month.upper()]
                month = month_no[month_en.upper()]

            date = datetime(int(year), int(month), int(day))

            time_str = timestamp.search(row).groups()[0]
            time = dateutil.parser.parse(time_str).time()

            date_time = datetime.combine(date, time)
        elif day_month_year.match(row) or year_month_day.match(row):
            if day_month_year.match(row):
                dayfirst = True
            else:
                dayfirst = False

            time_str = row_next

            date = dateutil.parser.parse(row, dayfirst=dayfirst).date()
            time = dateutil.parser.parse(time_str).time()

            date_time = datetime.combine(date, time)

        return date_time

    def tag_items(self, path):
        nmversion = re.compile(r'1NONLINEAR MIXED EFFECTS MODEL PROGRAM \(NONMEM\) VERSION\s+(\S+)')
        tag = re.compile(r'\s*#([A-Z]{4}):\s*(.*)')
        end_TERE = re.compile(r'(0|1)')  # The part we need after #TERE: will preceed next ^1 or ^0
        cleanup = re.compile(r'\*+\s*')
        TERM = list()
        TERE = list()
        found_TERM = False
        found_TERE = False
        found_runtime = False
        endtime_index = None

        with open(path, 'rb') as fp:
            binary = fp.readlines()
            # Since lst-files sometimes can have mixed encodings:
            # Allow for separate encodings of date strings and the rest of the content
            # Always try utf-8 first and fallback to latin-1 separately for each section
            try:
                line1 = binary[0].decode('utf-8')
            except UnicodeDecodeError:
                line1 = binary[0].decode('latin-1', errors='ignore')

            try:
                last_line = binary[-1].decode('utf-8')
            except UnicodeDecodeError:
                last_line = binary[-1].decode('latin-1', errors='ignore')

            chunk = b''.join(binary[1:-1])
            try:
                decoded_chunk = chunk.decode('utf-8')
            except UnicodeDecodeError:
                decoded_chunk = chunk.decode('latin-1', errors='ignore')

            lines = [line1]
            lines += decoded_chunk.split('\n')
            lines[-1] = last_line  # Replace since lst-files always end with \n

        version_number = None
        starttime = NONMEMResultsFile.parse_runtime(lines[0], lines[1])
        for row in lines:
            m = nmversion.match(row)
            if m:
                version_number = NONMEMResultsFile.cleanup_version(m.group(1))
                yield ('nonmem_version', version_number)
                break  # we will stay at current file position

        i = 0
        while i < len(lines):
            line = lines[i]
            if line.startswith('0WARNING:'):
                message = line[10:]
                if lines[i + 1].startswith(' '):
                    i += 1
                    message += lines[i]
                self.log.log_warning(message)
            elif line.startswith(" AN ERROR WAS FOUND IN THE CONTROL STATEMENTS."):
                if i + 6 < len(lines):
                    end = len(lines)
                else:
                    end = i + 6
                message = '\n'.join(lines[i:end])
                self.log.log_error(message)
                i += 5
            elif line.startswith('0PRED EXIT CODE = 1'):
                message = line[1:] + '\n' + lines[i + 1][1:].rstrip() + '\n'
                i += 1
                while not lines[i + 1].startswith('0PROGRAM TERMINATED'):
                    i += 1
                    message += lines[i].rstrip() + '\n'
                self.log.log_error(message)
            elif line.startswith('0PROGRAM TERMINATED'):
                message = line[1:]
                while i < len(lines):
                    i += 1
                    if lines[i].startswith(' '):
                        message += '\n' + lines[i]
                    else:
                        break
                self.log.log_error(message)
            i += 1

        if NONMEMResultsFile.supported_version(version_number):
            for i, row in enumerate(lines):
                row = row.rstrip()
                m = tag.match(row)
                if m:
                    if m.group(1) == 'TERM':
                        if found_TERM:
                            raise NotImplementedError('Two TERM tags without TERE in between')
                        found_TERM = True
                        TERM = list()
                    elif m.group(1) == 'TERE':
                        if not found_TERM:
                            raise NotImplementedError('TERE tag without TERM tag')
                        found_TERE = True
                        yield ('TERM', NONMEMResultsFile.parse_termination(TERM))
                        found_TERM = False
                        TERM = list()
                    elif found_TERE:
                        found_TERE = False
                        # raise NotImplementedError('TERE tag without ^1 or ^0 before next tag')
                    else:
                        v = cleanup.sub('', m.group(2))
                        yield (m.group(1), v.strip())
                elif found_TERE:
                    if end_TERE.match(row):
                        yield ('TERE', NONMEMResultsFile.parse_covariance(TERE))
                        found_TERE = False
                        TERE = list()
                    else:
                        TERE.append(row)
                elif found_TERM:
                    TERM.append(row)
                if row == 'Stop Time:':
                    endtime_index = i + 1

            if endtime_index is not None:
                if endtime_index + 1 < len(lines):
                    second_line = lines[endtime_index + 1]
                else:
                    second_line = None
                endtime = NONMEMResultsFile.parse_runtime(lines[endtime_index], second_line)
                if starttime and endtime:
                    runtime = (endtime - starttime).total_seconds()
                    found_runtime = True

            if found_TERM:
                yield ('TERM', NONMEMResultsFile.parse_termination(TERM))
            if found_TERE:
                yield ('TERE', NONMEMResultsFile.parse_covariance(TERE))
            if found_runtime:
                yield ('runtime', runtime)

    def table_blocks(self, path):
        block = dict()
        table_number = 'INIT'
        for name, content in self.tag_items(path):
            if name == 'TERM' or name == 'TERE':
                for k, v in content.items():
                    block[k] = v
            elif name == 'TBLN':
                if bool(block):
                    yield (table_number, block)
                block = dict()
                table_number = int(content)
            elif name == 'runtime':
                yield ('runtime', {'total': content})
            else:
                # If already set then it means TBLN was missing, probably $SIM, skip
                if name not in block.keys():
                    block[name] = content
        if bool(block):
            yield (table_number, block)



import datetime
from dataclasses import dataclass
from pathlib import Path

import pandas as pd


@dataclass
class LogEntry:
    category: str = None
    time: datetime.datetime = datetime.datetime.now()
    model: str = ""
    tool: str = ""
    task: str = ""
    path: Path = None
    message: str = ""


class Log:
    def __init__(self):
        self._log = []  # list of LogEntry

    def log_error(self, message):
        entry = LogEntry(category='ERROR', message=message)
        self._log.append(entry)

    def log_warning(self, message):
        entry = LogEntry(category='WARNING', message=message)
        self._log.append(entry)

    def to_dataframe(self):
        categories = [entry.category for entry in self._log]
        times = [entry.time for entry in self._log]
        models = [entry.model for entry in self._log]
        tools = [entry.tool for entry in self._log]
        tasks = [entry.task for entry in self._log]
        paths = [entry.path for entry in self._log]
        messages = [entry.message for entry in self._log]
        df = pd.DataFrame(
            {
                'category': categories,
                'time': times,
                'model': models,
                'tool': tools,
                'task': tasks,
                'path': paths,
                'message': messages,
            }
        )
        return df

    def __repr__(self):
        s = ''
        for entry in self._log:
            s += f'{entry.category:<8} {entry.time.strftime("%Y-%m-%d %H:%M:%S")}\n'
            for line in entry.message.split('\n'):
                s += f'    {line}\n'
        return s



import shutil

from pharmpy import Model
from pharmpy.modeling import run_tool
from pharmpy.utils import TemporaryDirectoryChanger


def test_amd(tmp_path, testdata):
    with TemporaryDirectoryChanger(tmp_path):
        for path in (testdata / 'nonmem').glob('pheno_real.*'):
            shutil.copy2(path, tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'sdtab1', tmp_path)

        model = Model('pheno_real.mod')
        model.dataset_path = tmp_path / 'pheno.dta'

        # FIXME: remove after updating results
        model.modelfit_results.estimation_step

        res = run_tool('amd', model)
        assert res



import shutil
from os import stat
from pathlib import Path

from pharmpy.model_factory import Model

from ..database import ModelDatabase, ToolDatabase


class LocalDirectoryToolDatabase(ToolDatabase):
    def __init__(self, toolname, path=None):
        if path is None:
            i = 1
            while True:
                name = f'{toolname}_dir{i}'
                path = Path(name)
                if not path.exists():
                    break
                i += 1
        path = Path(path)
        path.mkdir(parents=True)
        self.path = path.resolve()

        modeldb = LocalModelDirectoryDatabase(self.path / 'models')
        self.model_database = modeldb
        super().__init__(toolname)

    def store_local_file(self, source_path):
        if Path(source_path).is_file():
            shutil.copy2(source_path, self.path)


class LocalDirectoryDatabase(ModelDatabase):
    # Files are all stored in the same directory
    # Assuming filenames connected to a model are named modelname + extension
    def __init__(self, path='.', file_extension='.mod'):
        path = Path(path)
        if not path.exists():
            path.mkdir(parents=True)
        self.path = path.resolve()
        self.file_extension = file_extension

    def store_local_file(self, model, path):
        if Path(path).is_file():
            shutil.copy2(path, self.path)

    def retrieve_local_files(self, name, destination_path):
        # Retrieve all files stored for one model
        files = self.path.glob(f'{name}.*')
        for f in files:
            shutil.copy2(f, destination_path)

    def retrieve_file(self, model_name, filename):
        # Return path to file
        path = self.path / filename
        if path.is_file() and stat(path).st_size > 0:
            return path
        else:
            raise FileNotFoundError(f"Cannot retrieve {filename} for {model_name}")

    def get_model(self, name):
        filename = name + self.file_extension
        path = self.path / filename
        try:
            model = Model(path)
        except FileNotFoundError:
            raise KeyError('Model cannot be found in database')
        model.database = self
        model.read_modelfit_results()
        return model


class LocalModelDirectoryDatabase(LocalDirectoryDatabase):
    def store_local_file(self, model, path):
        if Path(path).is_file():
            destination = self.path / model.name
            if not destination.is_dir():
                destination.mkdir(parents=True)
            shutil.copy2(path, destination)

    def retrieve_local_files(self, name, destination_path):
        path = self.path / name
        files = path.glob('*')
        for f in files:
            shutil.copy2(f, destination_path)

    def retrieve_file(self, model_name, filename):
        # Return path to file
        path = self.path / model_name / filename
        if path.is_file() and stat(path).st_size > 0:
            return path
        else:
            raise FileNotFoundError(f"Cannot retrieve {filename} for {model_name}")

    def get_model(self, name):
        filename = name + self.file_extension
        path = self.path / name / filename
        model = Model(path)
        model.database = self
        model.read_modelfit_results()
        return model



import re

import pandas as pd

from pharmpy import Model
from pharmpy.results import Results


class CrossvalResults(Results):
    """Crossval results class"""

    def __init__(self, runs=None, prediction_ofv_sum=None):
        self.runs = runs
        self.prediction_ofv_sum = prediction_ofv_sum


def calculate_results(estimation_models, prediction_models):
    """Calculate crossval results"""
    est_ofvs = [model.modelfit_results.ofv for model in estimation_models]
    pred_ofvs = [model.modelfit_results.ofv for model in prediction_models]
    runs = pd.DataFrame(
        {'estimation_ofv': est_ofvs, 'prediction_ofv': pred_ofvs},
        index=pd.RangeIndex(start=1, stop=len(est_ofvs) + 1),
    )

    res = CrossvalResults(runs=runs, prediction_ofv_sum=sum(pred_ofvs))
    return res


def psn_crossval_results(path):
    def atoi(text):
        return int(text) if text.isdigit() else text

    def natural_keys(text):
        return [atoi(c) for c in re.split(r'(\d+)', text)]

    est_paths = [str(p) for p in (path / 'm1').glob('est_model*.mod')]
    est_paths.sort(key=natural_keys)
    pred_paths = [str(p) for p in (path / 'm1').glob('pred_model*.mod')]
    pred_paths.sort(key=natural_keys)
    ests = [Model(path) for path in est_paths]
    preds = [Model(path) for path in pred_paths]

    res = calculate_results(ests, preds)
    return res



import pytest


@pytest.mark.parametrize(
    'buf,title,new_title,new_str',
    [
        ('$PROB\n', '', 'PHENO  MODEL', '$PROB PHENO  MODEL\n'),
        ('$PROBLEM ABC\n', 'ABC', None, None),
        ('$PROBLEM   A;BC \n\n', 'A;BC', "New", "$PROBLEM New\n\n"),
        ('$PROBLEM A\n', 'A', None, None),
        ('$PROBLEM ABC \n', 'ABC', None, None),
        ('$PROBL A ; B ; C \n', 'A ; B ; C', None, None),
        ('$PROBL A ; B \n', 'A ; B', None, None),
        ('$PROBLEM \n ', '', None, None),
        ('$PROBLEM \n \n', '', None, None),
    ],
)
def test_modify_string(parser, buf, title, new_title, new_str):
    recs = parser.parse(buf)
    rec = recs.records[0]
    assert str(rec) == buf
    assert rec.title == title
    if new_title:
        rec.title = new_title
        assert rec.title == new_title
        assert str(rec) == new_str


def test_title_limits(parser):
    rec = parser.parse("$PROBLEM " + "k" * 72).records[0]
    assert rec.title == "k" * 72
    rec = parser.parse("$PROBLEM " + "k" * 73).records[0]
    assert rec.title == "k" * 72
    with pytest.raises(ValueError):
        rec.title = "k" * 73
    rec.title = "k" * 72
    rec.title = " mytitle"
    assert rec.title == "mytitle"



# The Dispatcher is the link between the local machine and the execution system.
# The run method should:
# 1. Create and cd into a temp directory
# 2. Copy the infiles from the local machine to the temp directory
# 3. Execute the workflow
# 4. Put the outfiles into the database


class ExecutionDispatcher:
    def run(self, workflow, database, infiles=None, outfiles=None):
        raise NotImplementedError()



from .database import ModelDatabase
from .databases import (
    LocalDirectoryDatabase,
    LocalDirectoryToolDatabase,
    LocalModelDirectoryDatabase,
    NullModelDatabase,
    NullToolDatabase,
)
from .dispatcher import ExecutionDispatcher
from .dispatchers import LocalDispatcher
from .execute import (
    default_dispatcher,
    default_model_database,
    default_tool_database,
    execute_workflow,
)
from .log import Log
from .workflows import Task, Workflow

__all__ = [
    'default_dispatcher',
    'default_model_database',
    'default_tool_database',
    'execute_workflow',
    'LocalDispatcher',
    'LocalDirectoryDatabase',
    'LocalModelDirectoryDatabase',
    'LocalDirectoryToolDatabase',
    'Log',
    'NullModelDatabase',
    'NullToolDatabase',
    'ExecutionDispatcher',
    'ModelDatabase',
    'Task',
    'Workflow',
]



import pandas as pd

import pharmpy.model
import pharmpy.tools
import pharmpy.tools.modelfit as modelfit
from pharmpy import Parameter, Parameters, RandomVariable, RandomVariables
from pharmpy.estimation import EstimationMethod
from pharmpy.modeling import set_iiv_on_ruv, set_power_on_ruv
from pharmpy.statements import Assignment, ModelStatements
from pharmpy.tools.modelfit import create_multiple_fit_workflow
from pharmpy.workflows import Task, Workflow

from .results import calculate_results


def create_workflow(model=None):
    wf = Workflow()
    wf.name = "resmod"  # FIXME: Could have as input to Workflow

    if model is not None:
        start_task = Task('start_resmod', start, model)
    else:
        start_task = Task('start_resmod', start)

    task_base_model = Task('create_base_model', _create_base_model)
    wf.add_task(task_base_model, predecessors=start_task)

    task_iiv = Task('create_iiv_on_ruv_model', _create_iiv_on_ruv_model)
    wf.add_task(task_iiv, predecessors=task_base_model)
    task_power = Task('create_power_model', _create_power_model)
    wf.add_task(task_power, predecessors=task_base_model)

    fit_wf = create_multiple_fit_workflow(n=3)
    wf.insert_workflow(fit_wf, predecessors=[task_base_model, task_iiv, task_power])

    task_post_process = Task('post_process', post_process)
    wf.add_task(task_post_process, predecessors=[start_task] + fit_wf.output_tasks)

    task_unpack = Task('unpack', _unpack)
    wf.add_task(task_unpack, predecessors=[task_post_process])

    fit_final = modelfit.create_single_fit_workflow()
    wf.insert_workflow(fit_final, predecessors=[task_unpack])

    task_results = Task('results', _results)
    wf.add_task(task_results, predecessors=[task_post_process] + fit_final.output_tasks)

    return wf


def start(model):
    return model


def post_process(start_model, *models):
    res = calculate_results(
        base_model=_find_model(models, 'base'),
        iiv_on_ruv=_find_model(models, 'iiv_on_ruv'),
        power=_find_model(models, 'power'),
    )
    best_model = _create_best_model(start_model, res)
    res.best_model = best_model
    return res


def _results(res, best_model):
    res.best_model = best_model
    return res


def _unpack(res):
    return res.best_model


def _find_model(models, name):
    for model in models:
        if model.name == name:
            return model


def _create_base_model(input_model):
    base_model = pharmpy.model.Model()
    theta = Parameter('theta', 0.1)
    omega = Parameter('omega', 0.01, lower=0)
    sigma = Parameter('sigma', 1, lower=0)
    params = Parameters([theta, omega, sigma])
    base_model.parameters = params

    eta = RandomVariable.normal('eta', 'iiv', 0, omega.symbol)
    sigma = RandomVariable.normal('epsilon', 'ruv', 0, sigma.symbol)
    rvs = RandomVariables([eta, sigma])
    base_model.random_variables = rvs

    y = Assignment('Y', theta.symbol + eta.symbol + sigma.symbol)
    stats = ModelStatements([y])
    base_model.statements = stats

    base_model.dependent_variable = y.symbol
    base_model.name = 'base'
    base_model.dataset = _create_dataset(input_model)
    base_model.database = input_model.database

    est = EstimationMethod('foce', interaction=True)
    base_model.estimation_steps = [est]
    return base_model


def _create_iiv_on_ruv_model(input_model):
    base_model = input_model
    model = base_model.copy()
    set_iiv_on_ruv(model)
    model.name = 'iiv_on_ruv'
    return model


def _create_power_model(input_model):
    base_model = input_model
    model = base_model.copy()
    set_power_on_ruv(model, ipred='IPRED')
    model.name = 'power'
    return model


def _create_dataset(input_model):
    residuals = input_model.modelfit_results.residuals
    cwres = residuals['CWRES']
    predictions = input_model.modelfit_results.predictions
    ipred = predictions['IPRED'].reindex(cwres.index)
    df = pd.concat([cwres, ipred], axis=1).rename(columns={'CWRES': 'DV'})
    df.reset_index(inplace=True)
    return df


def _create_best_model(model, res):
    model = model.copy()
    if any(-res.models['dofv'] > 3.84):
        idx = res.models['dofv'].idxmin()
        name = idx[0]
        if name == 'power':
            set_power_on_ruv(model)
            model.parameters.inits = {
                'power1': res.models['parameters'].loc['power', 1, 1].get('theta')
            }
        else:
            set_iiv_on_ruv(model)
            model.parameters.inits = {
                'IIV_RUV1': res.models['parameters'].loc['IIV_on_RUV', 1, 1].get('omega')
            }
        model.update_source()
    return model



"""

.. list-table:: Options for the results module
   :widths: 25 25 50 150
   :header-rows: 1

   * - Option name
     - Default value
     - Type
     - Description
   * - ``native_shrinkage``
     - ``True``
     - bool
     - Should shrinkage calculation of external tool be used.
       Otherwise pharmpy will calculate shrinkage
"""

import copy
import json
import lzma
import math
from collections.abc import MutableSequence
from pathlib import Path

import altair as alt
import numpy as np
import pandas as pd

import pharmpy.config as config
from pharmpy.math import cov2corr


class ResultsConfiguration(config.Configuration):
    module = 'pharmpy.results'
    """ FIXME: This setting should be moved to the NONMEM plugin
    """
    native_shrinkage = config.ConfigItem(True, 'Use shrinkage results from external tool')


conf = ResultsConfiguration()


class ResultsJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, pd.DataFrame) or isinstance(obj, pd.Series):
            d = json.loads(obj.to_json(orient='table'))
            if isinstance(obj, pd.DataFrame):
                d['__class__'] = 'DataFrame'
            elif isinstance(obj, pd.Series):
                d['__class__'] = 'Series'
            else:
                d['__class__'] = obj.__class__.__name__
            return d
        elif obj.__class__.__module__.startswith('altair.'):
            d = obj.to_dict()
            d['__class__'] = 'vega-lite'
            return d
        else:
            return json.JSONEncoder.encode(self, obj)


class ResultsJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)
        self.cls = None

    def object_hook(self, dct):
        if '__class__' in dct:
            cls = dct['__class__']
            del dct['__class__']
            if cls == 'DataFrame' or cls == 'Series':
                res = pd.read_json(json.dumps(dct), orient='table')
                return res
            elif cls == 'vega-lite':
                res = alt.Chart.from_dict(dct)
                return res
            else:
                self.cls = cls
        return dct


def read_results(path_or_buf):
    if '{' in str(path_or_buf):  # Heuristic to determine if path or buffer
        s = path_or_buf
    else:
        path = Path(path_or_buf)
        if path.is_dir():
            path /= 'results.json'
        if not path.is_file():
            raise FileNotFoundError(str(path))
        if path.name.endswith('.xz'):
            with lzma.open(path, 'r') as json_file:
                s = json_file.read().decode('utf-8')
        else:
            with open(path, 'r') as json_file:
                s = json_file.read()
    decoder = ResultsJSONDecoder()
    d = decoder.decode(s)
    if decoder.cls == 'FREMResults':
        from pharmpy.tools.frem import FREMResults

        res = FREMResults.from_dict(d)
    elif decoder.cls == 'BootstrapResults':
        from pharmpy.tools.bootstrap import BootstrapResults

        res = BootstrapResults.from_dict(d)
    elif decoder.cls == 'CDDResults':
        from pharmpy.tools.cdd import CDDResults

        res = CDDResults.from_dict(d)
    elif decoder.cls == 'SCMResults':
        from pharmpy.tools.scm import SCMResults

        res = SCMResults.from_dict(d)
    elif decoder.cls == 'QAResults':
        from pharmpy.tools.qa import QAResults

        res = QAResults.from_dict(d)
    elif decoder.cls == 'LinearizeResults':
        from pharmpy.tools.linearize import LinearizeResults

        res = LinearizeResults.from_dict(d)
    elif decoder.cls == 'ResmodResults':
        from pharmpy.tools.resmod import ResmodResults

        res = ResmodResults.from_dict(d)
    elif decoder.cls == 'SimevalResults':
        from pharmpy.tools.simeval import SimevalResults

        res = SimevalResults.from_dict(d)
    elif decoder.cls == 'CrossvalResults':
        from pharmpy.tools.crossval import CrossvalResults

        res = CrossvalResults.from_dict(d)

    return res


class Results:
    """Base class for all result classes"""

    @classmethod
    def from_dict(cls, d):
        return cls(**d)

    def to_json(self, path=None, lzma=False):
        json_dict = self.to_dict()
        json_dict['__class__'] = self.__class__.__name__
        s = json.dumps(json_dict, cls=ResultsJSONEncoder)
        if path:
            if not lzma:
                with open(path, 'w') as fh:
                    fh.write(s)
            else:
                xz_path = path.parent / (path.name + '.xz')
                with lzma.open(xz_path, 'w') as fh:
                    fh.write(bytes(s, 'utf-8'))
        else:
            return s

    def get_and_reset_index(self, attr, **kwargs):
        """Wrapper to reset index of attribute or result from method.

        Used to facilitate importing multiindex dataframes into R
        """
        val = getattr(self, attr)
        if callable(val):
            df = val(**kwargs)
        else:
            df = val
        return df.reset_index()

    def to_dict(self):
        """Convert results object to a dictionary"""
        return vars(self).copy()

    def __str__(self):
        start = self.__class__.__name__
        s = f'{start}\n\n'
        d = self.to_dict()
        for key, value in d.items():
            if value.__class__.__module__.startswith('altair.'):
                continue
            s += f'{key}\n'
            if isinstance(value, pd.DataFrame):
                s += value.to_string()
            elif isinstance(value, list):  # Print list of lists as table
                if len(value) > 0 and isinstance(value[0], list):
                    df = pd.DataFrame(value)
                    df_str = df.to_string(index=False)
                    df_str = df_str.split('\n')[1:]
                    s += '\n'.join(df_str)
            else:
                s += str(value) + '\n'
            s += '\n\n'
        return s

    def to_csv(self, path):
        """Save results as a human readable csv file

        Index will not be printed if it is a basic range.
        """
        d = self.to_dict()
        s = ""
        for key, value in d.items():
            if value.__class__.__module__.startswith('altair.'):
                continue
            s += f'{key}\n'
            if isinstance(value, pd.DataFrame):
                if isinstance(value.index, pd.RangeIndex):
                    use_index = False
                else:
                    use_index = True
                s += value.to_csv(index=use_index)
            elif isinstance(value, pd.Series):
                s += value.to_csv()
            elif isinstance(value, list):  # Print list of lists as table
                if len(value) > 0 and isinstance(value[0], list):
                    for row in value:
                        s += f'{",".join(map(str, row))}\n'
            else:
                s += str(value) + '\n'
            s += '\n'
        with open(path, 'w', newline='') as fh:
            print(s, file=fh)

    def create_report(self, path):
        import pharmpy.reporting.reporting as reporting

        reporting.generate_report(self.rst_path, path)

    def add_plots(self):
        """Create and add all plots to results object"""
        raise NotImplementedError()


class ModelfitResults(Results):
    """Base class for results from a modelfit operation

    model_name - name of model that generated the results model

    Attributes
    ----------
    correlation_matrix : pd.DataFrame
        Correlation matrix of the population parameter estimates
    covariance_matrix : pd.DataFrame
        Covariance matrix of the population parameter estimates
    information_matrix : pd.DataFrame
        Fischer information matrix of the population parameter estimates
    evaluation_ofv : float
        The objective function value as if the model was evaluated. Currently
        workfs for classical estimation methods by taking the OFV of the first
        iteration.
    individual_ofv : pd.Series
        OFV for each individual
    individual_estimates : pd.DataFrame
        Estimates for etas
    individual_estimates_covariance : pd.Series
        Estimated covariance between etas
    parameter_estimates : pd.Series
        Population parameter estimates
    parameter_estimates_sdcorr : pd.Series
        Population parameter estimates with variability parameters as standard deviations and
        correlations
    residuals: pd.DataFrame
        Table of various residuals
    runtime_total : float
        Total runtime of estimation
    standard_errors : pd.Series
        Standard errors of the population parameter estimates
    standard_errors_sdcorr : pd.Series
        Standard errors of the population parameter estimates on standard deviation and correlation
        scale
    """

    def __init__(
        self,
        ofv=None,
        parameter_estimates=None,
        parameter_estimates_sdcorr=None,
        covariance_matrix=None,
        correlation_matrix=None,
        information_matrix=None,
        standard_errors=None,
        minimization_successful=None,
        individual_ofv=None,
        individual_estimates=None,
        individual_estimates_covariance=None,
        residuals=None,
        runtime_total=None,
    ):
        self.ofv = ofv
        self.parameter_estimates = parameter_estimates
        self.parameter_estimates_sdcorr = parameter_estimates_sdcorr
        self.covariance_matrix = covariance_matrix
        self.correlation_matrix = correlation_matrix
        self.standard_errors = standard_errors
        self._minimization_successful = minimization_successful
        self.individual_estimates = individual_estimates
        self.individual_ofv = individual_ofv
        self.residuals = residuals
        self.runtime_total = runtime_total

    def __bool__(self):
        return bool(self.ofv) and bool(self.parameter_estimates)

    def to_dict(self):
        """Convert results object to a dictionary"""
        return {'parameter_estimates': self.parameter_estimates}

    @property
    def aic(self):
        """Final AIC value assuming the OFV to be -2LL"""
        parameters = self.model.parameters.copy()
        parameters.remove_fixed()
        return self.ofv + 2 * len(parameters)

    @property
    def bic(self):
        """Final BIC value assuming the OFV to be -2LL"""
        parameters = self.model.parameters.copy()
        parameters.remove_fixed()
        return self.ofv + len(parameters) * math.log(len(self.model.dataset.pharmpy.observations))

    @property
    def minimization_successful(self):
        """Was the minimization successful"""
        return self._minimization_successful

    def _cov_from_inf(self):
        Im = self.information_matrix
        cov = pd.DataFrame(np.linalg.inv(Im.values), index=Im.index, columns=Im.columns)
        return cov

    def _cov_from_corrse(self):
        se = self.standard_errors
        corr = self.correlation_matrix
        x = se.values @ se.values.T
        cov = x * corr.values
        cov_df = pd.DataFrame(cov, index=corr.index, columns=corr.columns)
        return cov_df

    def _inf_from_cov(self):
        C = self.covariance_matrix
        Im = pd.DataFrame(np.linalg.inv(C.values), index=C.index, columns=C.columns)
        return Im

    def _inf_from_corrse(self):
        se = self.standard_errors
        corr = self.correlation_matrix
        x = se.values @ se.values.T
        cov = x * corr.values
        Im = pd.DataFrame(np.linalg.inv(cov), index=corr.index, columns=corr.columns)
        return Im

    def _corr_from_cov(self):
        C = self.covariance_matrix
        corr = pd.DataFrame(cov2corr(C.values), index=C.index, columns=C.columns)
        return corr

    def _corr_from_inf(self):
        Im = self.information_matrix
        corr = pd.DataFrame(cov2corr(np.linalg.inv(Im.values)), index=Im.index, columns=Im.columns)
        return corr

    @property
    def relative_standard_errors(self):
        """Relative standard errors of population parameter estimates"""
        if self.standard_errors is not None:
            ser = self.standard_errors / self.parameter_estimates
            ser.name = 'RSE'
            return ser

    def _se_from_cov(self):
        """Calculate the standard errors from the covariance matrix
        can be used by subclasses
        """
        cov = self.covariance_matrix
        se = pd.Series(np.sqrt(np.diag(cov.values)), index=cov.index)
        return se

    def _se_from_inf(self):
        """Calculate the standard errors from the information matrix"""
        Im = self.information_matrix
        se = pd.Series(np.sqrt(np.linalg.inv(Im.values)), index=Im.index)
        return se

    def near_bounds(self, zero_limit=0.001, significant_digits=2):
        return self.model.parameters.is_close_to_bound(
            values=self.parameter_estimates,
            zero_limit=zero_limit,
            significant_digits=significant_digits,
        )

    def parameter_summary(self):
        """Summary of parameter estimates and uncertainty"""
        pe = self.parameter_estimates
        ses = self.standard_errors
        rses = self.relative_standard_errors
        df = pd.DataFrame({'estimate': pe, 'SE': ses, 'RSE': rses})
        return df

    def __repr__(self):
        df = self.parameter_summary()
        return df.to_string()


class ChainedModelfitResults(MutableSequence, ModelfitResults):
    """A sequence of modelfit results given in order from first to final
    inherits from both list and ModelfitResults. Each method from ModelfitResults
    will be performed on the final modelfit object
    """

    def __init__(self, results=None):
        if isinstance(results, ChainedModelfitResults):
            self._results = copy.deepcopy(results._results)
        elif results is None:
            self._results = []
        else:
            self._results = list(results)

    def __getitem__(self, ind):
        return self._results[ind]

    def __setitem__(self, ind, value):
        self._results[ind] = value

    def __delitem__(self, ind):
        del self._results[ind]

    def __len__(self):
        return len(self._results)

    def insert(self, ind, value):
        self._results.insert(ind, value)

    @property
    def ofv(self):
        return self[-1].ofv

    @property
    def evaluation_ofv(self):
        return self[0].evaluation_ofv

    @property
    def minimization_successful(self):
        # If last step was estimation
        est_steps = self.model.estimation_steps
        if not est_steps[-1].evaluation and self[-1].minimization_successful:
            return self[-1].minimization_successful
        # If last was evaluation, find latest estimation
        for i in reversed(range(len(self))):
            if not est_steps[i].evaluation:
                return self[i].minimization_successful
        # If all steps were evaluation the last evaluation step is relevant
        return self[-1].minimization_successful

    @property
    def parameter_estimates(self):
        return self[-1].parameter_estimates

    @parameter_estimates.setter
    def parameter_estimates(self, value):
        self[-1].parameter_estimates = value

    @property
    def parameter_estimates_sdcorr(self):
        return self[-1].parameter_estimates_sdcorr

    @property
    def covariance_matrix(self):
        return self[-1].covariance_matrix

    @property
    def information_matrix(self):
        return self[-1].information_matrix

    @property
    def correlation_matrix(self):
        return self[-1].correlation_matrix

    @property
    def standard_errors(self):
        return self[-1].standard_errors

    @property
    def standard_errors_sdcorr(self):
        return self[-1].standard_errors_sdcorr

    @property
    def individual_ofv(self):
        return self[-1].individual_ofv

    @property
    def individual_estimates(self):
        return self[-1].individual_estimates

    @property
    def individual_estimates_covariance(self):
        return self[-1].individual_estimates_covariance

    @property
    def residuals(self):
        return self[-1].residuals

    @property
    def predictions(self):
        return self[-1].predictions

    @property
    def model_name(self):
        return self[-1].model_name

    @property
    def runtime_total(self):
        return self[-1].runtime_total

    def result_summary(self, include_all_estimation_steps=False):
        if not include_all_estimation_steps:
            summary_dict = self._summarize_step(-1)
            summary_df = pd.DataFrame(summary_dict, index=[self.model_name])
            return summary_df
        else:
            summary_dicts = []
            tuples = []
            for i in range(len(self)):
                summary_dict = self._summarize_step(i)
                is_evaluation = self.model.estimation_steps[i].evaluation
                if is_evaluation:
                    run_type = 'evaluation'
                else:
                    run_type = 'estimation'
                summary_dict = {**{'run_type': run_type}, **summary_dict}
                summary_dicts.append(summary_dict)
                tuples.append((self.model_name, i + 1))
            index = pd.MultiIndex.from_tuples(tuples, names=['model_name', 'step'])
            summary_df = pd.DataFrame(summary_dicts, index=index)
            return summary_df

    def _summarize_step(self, i):
        summary_dict = dict()

        if i >= 0:
            step = self[i]
        else:
            step = self

        summary_dict['minimization_successful'] = step.minimization_successful
        summary_dict['ofv'] = step.ofv
        summary_dict['runtime_total'] = step.runtime_total

        pe = step.parameter_estimates
        ses = step.standard_errors
        rses = step.relative_standard_errors

        for param in pe.index:
            summary_dict[f'{param}_estimate'] = pe[param]
            if ses is not None:
                summary_dict[f'{param}_SE'] = ses[param]
            if rses is not None:
                summary_dict[f'{param}_RSE'] = rses[param]

        return summary_dict

    def __repr__(self):
        return repr(self._results[-1])

    # FIXME: To not have to manually intercept everything here. Could do it in a general way.



import numpy as np
import pandas.testing
import pytest

import pharmpy.data
import pharmpy.data.iterators as iters


@pytest.fixture
def df():
    return pharmpy.data.PharmDataFrame(
        {'ID': [1, 1, 2, 2, 4, 4], 'DV': [5, 6, 3, 4, 0, 9], 'STRAT': [1, 1, 2, 2, 2, 2]}
    )


def test_omit(df):
    omitter = iters.Omit(df, 'ID')
    (new_df, group) = next(omitter)
    assert group == 1
    assert list(new_df['ID']) == [2, 2, 4, 4]
    assert list(new_df['DV']) == [3, 4, 0, 9]
    assert new_df.name == 'omitted_1'
    assert not hasattr(df, 'name')  # Check that original did not get name
    (new_df, group) = next(omitter)
    assert group == 2
    assert list(new_df['ID']) == [1, 1, 4, 4]
    assert list(new_df['DV']) == [5, 6, 0, 9]
    assert new_df.name == 'omitted_2'
    (new_df, group) = next(omitter)
    assert group == 4
    assert list(new_df['ID']) == [1, 1, 2, 2]
    assert list(new_df['DV']) == [5, 6, 3, 4]
    assert new_df.name == 'omitted_3'
    with pytest.raises(StopIteration):
        next(omitter)


def test_resampler_default(df):
    np.random.seed(28)
    resampler = iters.Resample(df, 'ID')
    (new_df, ids) = next(resampler)
    assert ids == [1, 4, 2]
    assert list(new_df['ID']) == [1, 1, 2, 2, 3, 3]
    assert list(new_df['DV']) == [5, 6, 0, 9, 3, 4]
    assert list(new_df['STRAT']) == [1, 1, 2, 2, 2, 2]
    assert new_df.name == 'resample_1'
    with pytest.raises(StopIteration):  # Test the default one iteration
        next(resampler)


def test_resampler_too_big_sample_size(df):
    with pytest.raises(ValueError):
        iters.Resample(df, 'ID', replace=False, sample_size=4)


def test_resampler_noreplace(df):
    np.random.seed(28)
    resampler = iters.Resample(df, 'ID', replace=False, sample_size=3)
    next(resampler)

    resampler = iters.Resample(df, 'ID', stratify='STRAT')
    (new_df, ids) = next(resampler)
    assert ids == [1, 2, 4]
    assert list(new_df['ID']) == [1, 1, 2, 2, 3, 3]
    assert list(new_df['DV']) == [5, 6, 3, 4, 0, 9]

    resampler = iters.Resample(df, 'ID', replace=False, sample_size=2)
    (new_df, ids) = next(resampler)
    assert list(ids) == [2, 1]
    assert list(new_df['ID']) == [1, 1, 2, 2]


def test_stratification(df):
    np.random.seed(28)
    resampler = iters.Resample(
        df, 'ID', resamples=1, stratify='STRAT', sample_size={1: 2, 2: 3}, replace=True
    )
    (new_df, ids) = next(resampler)
    assert ids == [1, 1, 4, 4, 4]
    assert list(new_df['ID']) == [1, 1, 2, 2, 3, 3, 4, 4, 5, 5]
    assert list(new_df['DV']) == [5, 6, 5, 6, 0, 9, 0, 9, 0, 9]

    resampler = iters.Resample(
        df, 'ID', resamples=1, stratify='STRAT', sample_size={1: 2}, replace=True
    )
    (new_df, ids) = next(resampler)
    assert ids == [1, 1]

    resampler = iters.Resample(df, 'ID', resamples=3, stratify='STRAT', replace=True)
    (new_df, ids) = next(resampler)
    assert ids == [1, 2, 2]
    (new_df, ids) = next(resampler)
    assert ids == [1, 2, 4]
    (new_df, ids) = next(resampler)
    assert ids == [1, 4, 2]


def test_resampler_anonymization(testdata):
    np.random.seed(28)
    df = pharmpy.data.read_csv(testdata / 'pheno_data.csv')
    resampler = iters.Resample(df, group='ID')
    (new_df, ids) = next(resampler)
    assert all(e in ids for e in range(1, 60))
    assert len(ids) == 59
    for new_id, old_id in enumerate(ids, start=1):
        df_newid = new_df.query(f'ID == {new_id}')
        df_oldid = df.query(f'ID == {old_id}')
        df_newid = df_newid.reset_index(drop=True)
        df_oldid.reset_index(inplace=True, drop=True)
        df_newid['ID'] = old_id
        pandas.testing.assert_frame_equal(df_newid, df_oldid)



from io import StringIO

from pharmpy import Model, RandomVariable, RandomVariables
from pharmpy.tools.iiv.algorithms import _get_iiv_combinations
from pharmpy.tools.iiv.tool import _get_iiv_block


def test_get_iiv_combinations(testdata, pheno_path):
    model = Model(
        StringIO(
            '''
$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2

$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
S1=V*EXP(ETA(3))*EXP(ETA(4))

$ERROR
Y=F+F*EPS(1)

$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA DIAGONAL(4)
 0.0309626  ;       IVCL
 0.031128  ;        IVV
 0.031128
 0.031128
$SIGMA 0.013241

$ESTIMATION METHOD=1 INTERACTION
'''
        )
    )

    model.path = testdata / 'nonmem' / 'pheno.mod'  # To be able to find dataset

    assert _get_iiv_combinations(model) == [
        ['ETA(1)', 'ETA(2)'],
        ['ETA(1)', 'ETA(3)'],
        ['ETA(1)', 'ETA(4)'],
        ['ETA(2)', 'ETA(3)'],
        ['ETA(2)', 'ETA(4)'],
        ['ETA(3)', 'ETA(4)'],
        ['ETA(1)', 'ETA(2)', 'ETA(3)'],
        ['ETA(1)', 'ETA(2)', 'ETA(4)'],
        ['ETA(1)', 'ETA(3)', 'ETA(4)'],
        ['ETA(2)', 'ETA(3)', 'ETA(4)'],
        ['ETA(1)', 'ETA(2)', 'ETA(3)', 'ETA(4)'],
    ]


def test_get_iiv_block():
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, 1)
    rv2, rv3 = RandomVariable.joint_normal(
        ['ETA(2)', 'ETA(3)'], 'iiv', [0, 0], [[1, 0.1], [0.1, 2]]
    )
    rv4 = RandomVariable.normal('ETA(4)', 'iiv', 0, 1)
    rvs = RandomVariables([rv1, rv2, rv3, rv4])
    assert _get_iiv_block(rvs) == ['ETA(2)', 'ETA(3)']



import shutil

import pytest

from pharmpy import Model
from pharmpy.modeling import add_iiv, run_tool
from pharmpy.utils import TemporaryDirectoryChanger


@pytest.mark.parametrize(
    'list_of_parameters, no_of_models',
    [([], 1), (['S1'], 4)],
)
def test_iiv(tmp_path, testdata, list_of_parameters, no_of_models):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'pheno_real.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path)
        model = Model('pheno_real.mod')
        model.dataset_path = tmp_path / 'pheno.dta'

        add_iiv(model, list_of_parameters, 'add')

        res = run_tool('iiv', 'brute_force', model=model)

        assert len(res.summary) == no_of_models
        assert len(res.models) == no_of_models
        assert all(int(model.modelfit_results.ofv) in range(570, 590) for model in res.models)
        rundir = tmp_path / 'iiv_dir1'
        assert rundir.is_dir()
        assert len(list((rundir / 'models').iterdir())) == no_of_models + 1



from pharmpy import Model
from pharmpy.modeling import (
    get_number_of_individuals,
    get_number_of_observations,
    get_number_of_observations_per_individual,
    get_observations,
)


def test_get_number_of_individuals(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_real.mod')
    assert get_number_of_individuals(model) == 59


def test_number_of_observations(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_real.mod')
    assert get_number_of_observations(model) == 155
    assert list(get_number_of_observations_per_individual(model)) == [
        2,
        3,
        3,
        3,
        3,
        3,
        3,
        3,
        4,
        3,
        1,
        3,
        2,
        4,
        2,
        3,
        3,
        4,
        3,
        3,
        3,
        2,
        3,
        3,
        6,
        2,
        2,
        1,
        1,
        2,
        1,
        3,
        2,
        2,
        2,
        3,
        2,
        4,
        3,
        2,
        3,
        2,
        1,
        3,
        3,
        1,
        1,
        5,
        3,
        4,
        3,
        3,
        2,
        4,
        1,
        1,
        2,
        3,
        3,
    ]


def test_get_observations(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_real.mod')
    obs = get_observations(model)
    assert len(obs) == 155
    assert obs.loc[1, 2.0] == 17.3



from pharmpy.parse_utils import AttrTree

from .record import Record


class ProblemRecord(Record):
    @property
    def title(self):
        first_line = str(self.root.raw_title).split('\n')[0].strip()
        max_number_of_characters_in_title = 72
        return first_line[:max_number_of_characters_in_title]

    @title.setter
    def title(self, new_title):
        if len(new_title.strip()) > 72:
            raise ValueError(
                f'The provided title is {len(new_title.strip())} characters long but '
                f'can be maximum 72 characters long (not counting initial and trailing'
                f'whitespace)'
            )
        if new_title == new_title.lstrip():
            new_title = ' ' + new_title
        lines = str(self.root.raw_title).split('\n')
        lines[0] = new_title
        s = '\n'.join(lines)
        node = AttrTree.create('raw_title', dict(ANYTHING=s))
        self.root.set('raw_title', node)



from io import StringIO

from pharmpy import Model


def test_des(testdata):
    code = """$PROBLEM
$INPUT ID TIME DV AMT
$DATA data.csv IGNORE=@
$SUBROUTINES ADVAN13 TOL=6
$MODEL COMP=(DOSE) COMP=(CENTRAL) COMP=(PERIPH) COMP=(LIVER)
$PK
MAT = THETA(1) * EXP(ETA(1))
CL = THETA(2) * EXP(ETA(2))
VC = THETA(3) * EXP(ETA(3))
Q = THETA(4) * EXP(ETA(4))
VP = THETA(5) * EXP(ETA(5))
VM = THETA(6) * EXP(ETA(6))
KM = THETA(7) * EXP(ETA(7))
KLO = THETA(8) * EXP(ETA(8))
VL = THETA(9) * EXP(ETA(9))
KA = 1/MAT
$DES
DADT(1) = -(KA * A(1))
DADT(2) = KA * A(1) + Q/VP * A(3) - (Q/VC * A(2) + A(2)/VC * VM/(KM + A(2)/VC))
DADT(3) = Q/VC * A(2) - Q/VP * A(3)
DADT(4) = A(2)/VC * VM/(KM + A(2)/VC) - KLO * A(4)
$ERROR
CONC = A(2)/VC
Y = CONC + EPS(1)
$ESTIMATION METHOD=COND INTER
$COVARIANCE PRINT=E
$THETA (0, 1, Inf) ; POP_MAT
$THETA (0, 1, Inf) ; POP_CL
$THETA (0, 1, Inf) ; POP_VC
$THETA (0, 1, Inf) ; POP_Q
$THETA (0, 1, Inf) ; POP_VP
$THETA (0, 1, Inf) ; POP_VM
$THETA (0, 1, Inf) ; POP_KM
$THETA (0, 1, Inf) ; POP_KLO
$THETA (0, 1, Inf) ; POP_VL
$OMEGA 0.1; IIV_MAT
$OMEGA 0.1; IIV_CL
$OMEGA 0.1; IIV_VC
$OMEGA 0.1; IIV_Q
$OMEGA 0.1; IIV_VP
$OMEGA 0.1; IIV_VM
$OMEGA 0.1; IIV_KM
$OMEGA 0.1; IIV_KLO
$OMEGA 0.1; IIV_VL
$SIGMA 0.1; RUV_ADD
"""
    pheno = Model(testdata / 'nonmem' / 'pheno.mod')
    model = Model(StringIO(code))
    model.dataset = pheno.dataset
    cs = model.statements.ode_system.to_compartmental_system()
    assert len(cs) == 5



"""
:meta private:
"""

import sympy

import pharmpy.symbols
from pharmpy.model import ModelException
from pharmpy.modeling.help_functions import _as_integer
from pharmpy.parameter import Parameter
from pharmpy.statements import Assignment, Bolus, CompartmentalSystem, Infusion


def _extract_params_from_symb(statements, symbol_name, pset):
    terms = {
        symb.name
        for symb in statements.before_odes.full_expression(sympy.Symbol(symbol_name)).free_symbols
    }
    theta_name = terms.intersection(pset.names).pop()
    return pset[theta_name]


def add_individual_parameter(model, name):
    """Add an individual or pk parameter to a model

    Parameters
    ----------
    model : Model
        Pharmpy model
    name : str
        Name of individual/pk parameter

    Return
    ------
    Model
        Reference to same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> add_individual_parameter(model, "KA")   # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("KA")
    KA := POP_KA

    """
    _add_parameter(model, name)
    return model


def _add_parameter(model, name, init=0.1):
    pops = model.create_symbol(f'POP_{name}')
    pop_param = Parameter(pops.name, init=init, lower=0)
    model.parameters.append(pop_param)
    symb = model.create_symbol(name)
    ass = Assignment(symb, pop_param.symbol)
    model.statements.insert(0, ass)
    return symb


def set_first_order_elimination(model):
    """Sets elimination to first order

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Reference to same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> set_first_order_elimination(model)     # doctest: +ELLIPSIS
    <...>
    >>> model.statements.ode_system
    Bolus(AMT)
    ┌───────┐       ┌──────┐
    │CENTRAL│──CL/V→│OUTPUT│
    └───────┘       └──────┘

    See also
    --------
    set_zero_order_elimination
    set_michaelis_menten_elimination

    """
    return model


def set_zero_order_elimination(model):
    """Sets elimination to zero order.

    Initial estimate for KM is set to 1% of smallest observation.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Reference to same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> set_zero_order_elimination(model)     # doctest: +ELLIPSIS
    <...>
    >>> model.statements.ode_system
    Bolus(AMT)
    ┌───────┐                                    ┌──────┐
    │CENTRAL│──CLMM*KM/(V*(KM + A_CENTRAL(t)/V))→│OUTPUT│
    └───────┘                                    └──────┘

    See also
    --------
    set_first_order_elimination
    set_michaelis_menten_elimination

    """
    _do_michaelis_menten_elimination(model)
    obs = model.dataset.pharmpy.observations
    init = obs.min() / 100  # 1% of smallest observation
    model.parameters['POP_KM'].init = init
    model.parameters['POP_KM'].fix = True
    return model


def set_michaelis_menten_elimination(model):
    """Sets elimination to Michaelis-Menten.

    Initial estimate for CLMM is set to CL and KM is set to :math:`2*max(DV)`.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> set_michaelis_menten_elimination(model)     # doctest: +ELLIPSIS
    <...>
    >>> model.statements.ode_system
    Bolus(AMT)
    ┌───────┐                                    ┌──────┐
    │CENTRAL│──CLMM*KM/(V*(KM + A_CENTRAL(t)/V))→│OUTPUT│
    └───────┘                                    └──────┘

    See also
    --------
    set_first_order_elimination
    set_zero_order_elimination

    """
    _do_michaelis_menten_elimination(model)
    return model


def set_mixed_mm_fo_elimination(model):
    """Sets elimination to mixed Michaelis-Menten and first order.

    Initial estimate for CLMM is set to CL/2 and KM is set to :math:`2*max(DV)`.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> set_mixed_mm_fo_elimination(model)     # doctest: +ELLIPSIS
    <...>
    >>> model.statements.ode_system
    Bolus(AMT)
    ┌───────┐                                         ┌──────┐
    │CENTRAL│──(CL + CLMM*KM/(KM + A_CENTRAL(t)/V))/V→│OUTPUT│
    └───────┘                                         └──────┘

    See also
    --------
    set_first_order_elimination
    set_zero_order_elimination
    set_michaelis_menten_elimination

    """
    _do_michaelis_menten_elimination(model, combined=True)
    return model


def _do_michaelis_menten_elimination(model, combined=False):
    odes = model.statements.ode_system
    central = odes.central_compartment
    output = odes.output_compartment
    old_rate = odes.get_flow(central, output)
    numer, denom = old_rate.as_numer_denom()

    km_init, clmm_init = _get_mm_inits(model, numer, combined)

    km = _add_parameter(model, 'KM', init=km_init)
    clmm = _add_parameter(model, 'CLMM', init=clmm_init)

    if denom != 1:
        if combined:
            cl = numer
        vc = denom
    else:
        if combined:
            cl = _add_parameter(model, 'CL', clmm_init)
        vc = _add_parameter(model, 'VC')  # FIXME: decide better initial estimate
    if not combined:
        cl = 0

    amount = sympy.Function(central.amount.name)(pharmpy.symbols.symbol('t'))
    rate = (clmm * km / (km + amount / vc) + cl) / vc
    odes.add_flow(central, output, rate)
    model.statements.remove_symbol_definitions(numer.free_symbols, odes)
    model.remove_unused_parameters_and_rvs()
    return model


def _get_mm_inits(model, rate_numer, combined):
    pset, sset = model.parameters, model.statements
    clmm_init = _extract_params_from_symb(sset, rate_numer.name, pset).init

    if combined:
        clmm_init /= 2

    dv_max = model.dataset.pharmpy.observations.max()
    km_init = dv_max * 2

    return km_init, clmm_init


def set_transit_compartments(model, n):
    """Set the number of transit compartments of model.

    Initial estimate for absorption rate is
    set the previous rate if available, otherwise it is set to the time of first observation/2.

    Parameters
    ----------
    model : Model
        Pharmpy model
    n : int
        Number of transit compartments

    Return
    ------
    Model
        Reference to same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> set_transit_compartments(model, 3)     # doctest: +ELLIPSIS
    <...>
    >>> model.statements.ode_system
    Bolus(AMT)
    ┌────────┐        ┌────────┐        ┌────────┐        ┌───────┐       ┌──────┐
    │TRANSIT1│──3/MDT→│TRANSIT2│──3/MDT→│TRANSIT3│──3/MDT→│CENTRAL│──CL/V→│OUTPUT│
    └────────┘        └────────┘        └────────┘        └───────┘       └──────┘

    See also
    --------
    add_lag_time

    """
    statements = model.statements
    odes = statements.ode_system
    transits = odes.find_transit_compartments(statements)
    try:
        n = _as_integer(n)
    except ValueError:
        raise ValueError(f'Number of compartments must be integer: {n}')
    if len(transits) == n:
        pass
    elif len(transits) == 0:
        mdt_symb = _add_parameter(model, 'MDT', init=_get_absorption_init(model, 'MDT'))
        rate = n / mdt_symb
        comp = odes.dosing_compartment
        dose = comp.dose
        comp.dose = None
        while n > 0:
            new_comp = odes.add_compartment(f'TRANSIT{n}')
            n -= 1
            odes.add_flow(new_comp, comp, rate)
            comp = new_comp
        comp.dose = dose
    elif len(transits) > n:
        nremove = len(transits) - n
        removed_symbols = set()
        trans, destination, flow = _find_last_transit(odes, transits)
        if n == 0:  # The dosing compartment will be removed
            dosing = odes.dosing_compartment
            dose = dosing.dose
        remaining = set(transits)
        while nremove > 0:
            from_comp, from_flow = odes.get_compartment_inflows(trans)[0]
            odes.add_flow(from_comp, destination, from_flow)
            odes.remove_compartment(trans)
            remaining.remove(trans)
            removed_symbols |= flow.free_symbols
            trans = from_comp
            flow = from_flow
            nremove -= 1
        if n == 0:
            destination.dose = dose
        _update_numerators(model)
        statements.remove_symbol_definitions(removed_symbols, odes)
        model.remove_unused_parameters_and_rvs()
    else:
        nadd = n - len(transits)
        last, destination, rate = _find_last_transit(odes, transits)
        odes.remove_flow(last, destination)
        while nadd > 0:
            # new_comp = odes.add_compartment(f'TRANSIT{len(transits) + nadd}')
            new_comp = odes.add_compartment(f'TRANSIT{n - nadd + 1}')
            odes.add_flow(last, new_comp, rate)
            if rate.is_Symbol:
                ass = statements.find_assignment(rate.name)
                if ass is not None:
                    rate = ass.expression
            last = new_comp
            nadd -= 1
        odes.add_flow(last, destination, rate)
        _update_numerators(model)
    return model


def _find_last_transit(odes, transits):
    for trans in transits:
        destination, flow = odes.get_compartment_outflows(trans)[0]
        if destination not in transits:
            return trans, destination, flow


def _update_numerators(model):
    # update numerators for transit compartment rates
    statements = model.statements
    odes = statements.ode_system
    transits = odes.find_transit_compartments(statements)
    new_numerator = sympy.Integer(len(transits))
    for comp in transits:
        to_comp, rate = odes.get_compartment_outflows(comp)[0]
        numer, denom = rate.as_numer_denom()
        if numer.is_Integer and numer != new_numerator:
            new_rate = new_numerator / denom
            odes.add_flow(comp, to_comp, new_rate)
        elif numer.is_Symbol:
            ass = statements.find_assignment(numer.name)
            if ass is not None:
                ass_numer, ass_denom = ass.expression.as_numer_denom()
                if ass_numer.is_Integer and ass_numer != new_numerator:
                    new_rate = new_numerator / ass_denom
                    statements.reassign(numer, new_rate)


def add_lag_time(model):
    """Add lag time to the dose compartment of model.

    Initial estimate for lag time is set the
    previous lag time if available, otherwise it is set to the time of first observation/2.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Reference to same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> add_lag_time(model)     # doctest: +ELLIPSIS
    <...>

    See also
    --------
    set_transit_compartments
    remove_lag_time

    """
    odes = model.statements.ode_system
    dosing_comp = odes.dosing_compartment
    old_lag_time = dosing_comp.lag_time
    mdt_symb = _add_parameter(model, 'MDT', init=_get_absorption_init(model, 'MDT'))
    dosing_comp.lag_time = mdt_symb
    if old_lag_time:
        model.statements.remove_symbol_definitions(old_lag_time.free_symbols, odes)
        model.remove_unused_parameters_and_rvs()
    return model


def remove_lag_time(model):
    """Remove lag time from the dose compartment of model.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Reference to same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> remove_lag_time(model)     # doctest: +ELLIPSIS
    <...>

    See also
    --------
    set_transit_compartments
    add_lag_time


    """
    odes = model.statements.ode_system
    dosing_comp = odes.dosing_compartment
    lag_time = dosing_comp.lag_time
    if lag_time:
        symbols = lag_time.free_symbols
        dosing_comp.lag_time = 0
        model.statements.remove_symbol_definitions(symbols, odes)
        model.remove_unused_parameters_and_rvs()
    return model


def set_zero_order_absorption(model):
    """Set or change to zero order absorption rate.

    Initial estimate for absorption rate is set
    the previous rate if available, otherwise it is set to the time of first observation/2.

    Parameters
    ----------
    model : Model
        Model to set or change to first order absorption rate

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> set_zero_order_absorption(model)     # doctest: +ELLIPSIS
    <...>
    >>> model.statements.ode_system
    Infusion(AMT, duration=2*MAT)
    ┌───────┐       ┌──────┐
    │CENTRAL│──CL/V→│OUTPUT│
    └───────┘       └──────┘

    See also
    --------
    set_bolus_order_absorption
    set_first_order_absorption

    """
    statements = model.statements
    odes = statements.ode_system
    if not isinstance(odes, CompartmentalSystem):
        raise ValueError("Setting absorption is not supported for ExplicitODESystem")
    _disallow_infusion(model)
    depot = odes.find_depot(statements)

    dose_comp = odes.dosing_compartment
    symbols = dose_comp.free_symbols
    dose = dose_comp.dose
    if depot:
        to_comp, _ = odes.get_compartment_outflows(depot)[0]
        ka = odes.get_flow(depot, odes.central_compartment)
        odes.remove_compartment(depot)
        symbols |= ka.free_symbols
        to_comp.dose = dose
    else:
        to_comp = dose_comp
    statements.remove_symbol_definitions(symbols, odes)
    model.remove_unused_parameters_and_rvs()
    if not has_zero_order_absorption(model):
        _add_zero_order_absorption(model, dose.amount, to_comp, 'MAT')
    return model


def set_first_order_absorption(model):
    """Set or change to first order absorption rate.

    Initial estimate for absorption rate is set to
    the previous rate if available, otherwise it is set to the time of first observation/2.

    Parameters
    ----------
    model : Model
        Model to set or change to use first order absorption rate

    Return
    ------
    Model
        Reference to same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> set_first_order_absorption(model)     # doctest: +ELLIPSIS
    <...>
    >>> model.statements.ode_system
    Bolus(AMT)
    ┌─────┐        ┌───────┐       ┌──────┐
    │DEPOT│──1/MAT→│CENTRAL│──CL/V→│OUTPUT│
    └─────┘        └───────┘       └──────┘

    See also
    --------
    set_bolus_order_absorption
    set_zero_order_absorption

    """
    statements = model.statements
    statements.to_compartmental_system()
    odes = statements.ode_system
    depot = odes.find_depot(statements)

    dose_comp = odes.dosing_compartment
    amount = dose_comp.dose.amount
    symbols = dose_comp.free_symbols
    if depot and depot == dose_comp:
        dose_comp.dose = Bolus(dose_comp.dose.amount)
    elif not depot:
        dose_comp.dose = None
    statements.remove_symbol_definitions(symbols, odes)
    model.remove_unused_parameters_and_rvs()
    if not depot:
        _add_first_order_absorption(model, Bolus(amount), dose_comp)
    return model


def set_bolus_absorption(model):
    """Set or change to bolus absorption rate.

    Parameters
    ----------
    model : Model
        Model to set or change absorption rate

    Return
    ------
    Model
        Reference to same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> set_bolus_absorption(model)     # doctest: +ELLIPSIS
    <...>
    >>> model.statements.ode_system
    Bolus(AMT)
    ┌───────┐       ┌──────┐
    │CENTRAL│──CL/V→│OUTPUT│
    └───────┘       └──────┘

    See also
    --------
    set_zero_order_absorption
    set_first_order_absorption

    """
    statements = model.statements
    odes = statements.ode_system
    if not isinstance(odes, CompartmentalSystem):
        raise ValueError("Setting absorption is not supported for ExplicitODESystem")
    depot = odes.find_depot(statements)
    if depot:
        to_comp, _ = odes.get_compartment_outflows(depot)[0]
        to_comp.dose = depot.dose
        ka = odes.get_flow(depot, odes.central_compartment)
        odes.remove_compartment(depot)
        symbols = ka.free_symbols
        statements.remove_symbol_definitions(symbols, odes)
        model.remove_unused_parameters_and_rvs()
    if has_zero_order_absorption(model):
        dose_comp = odes.dosing_compartment
        old_symbols = dose_comp.free_symbols
        dose_comp.dose = Bolus(dose_comp.dose.amount)
        unneeded_symbols = old_symbols - dose_comp.dose.free_symbols
        statements.remove_symbol_definitions(unneeded_symbols, odes)
        model.remove_unused_parameters_and_rvs()
    return model


def set_seq_zo_fo_absorption(model):
    """Set or change to sequential zero order first order absorption rate.

    Initial estimate for
    absorption rate is set the previous rate if available, otherwise it is set to the time of
    first observation/2.

    Parameters
    ----------
    model : Model
        Model to set or change absorption rate

    Return
    ------
    Model
        Reference to same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> set_seq_zo_fo_absorption(model)     # doctest: +ELLIPSIS
    <...>
    >>> model.statements.ode_system
    Infusion(AMT, duration=2*MDT)
    ┌─────┐        ┌───────┐       ┌──────┐
    │DEPOT│──1/MAT→│CENTRAL│──CL/V→│OUTPUT│
    └─────┘        └───────┘       └──────┘

    See also
    --------
    set_bolus_order_absorption
    set_zero_order_absorption
    set_first_order_absorption

    """
    statements = model.statements
    odes = statements.ode_system
    if not isinstance(odes, CompartmentalSystem):
        raise ValueError("Setting absorption is not supported for ExplicitODESystem")
    _disallow_infusion(model)
    depot = odes.find_depot(statements)

    dose_comp = odes.dosing_compartment
    have_ZO = has_zero_order_absorption(model)
    if depot and not have_ZO:
        _add_zero_order_absorption(model, dose_comp.amount, depot, 'MDT')
    elif not depot and have_ZO:
        _add_first_order_absorption(model, dose_comp.dose, dose_comp)
        dose_comp.dose = None
    elif not depot and not have_ZO:
        amount = dose_comp.dose.amount
        dose_comp.dose = None
        depot = _add_first_order_absorption(model, amount, dose_comp)
        _add_zero_order_absorption(model, amount, depot, 'MDT')
    return model


def _disallow_infusion(model):
    odes = model.statements.ode_system
    dose_comp = odes.dosing_compartment
    if isinstance(dose_comp.dose, Infusion):
        if dose_comp.dose.rate is not None:
            ex = dose_comp.dose.rate
        else:
            ex = dose_comp.dose.duration
        for s in ex.free_symbols:
            if s.name in model.dataset.columns:
                raise ModelException("Model already has an infusion given in the dataset")


def has_zero_order_absorption(model):
    """Check if ode system describes a zero order absorption

    currently defined as having Infusion dose with rate not in dataset

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Reference to same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> has_zero_order_absorption(model)
    False

    """
    odes = model.statements.ode_system
    dosing = odes.dosing_compartment
    dose = dosing.dose
    if isinstance(dose, Infusion):
        if dose.rate is None:
            value = dose.duration
        else:
            value = dose.rate
        if isinstance(value, sympy.Symbol) or isinstance(value, str):
            name = str(value)
            if name not in model.dataset.columns:
                return True
    return False


def _add_zero_order_absorption(model, amount, to_comp, parameter_name):
    """Add zero order absorption to a compartment. Initial estimate for absorption rate is set
    the previous rate if available, otherwise it is set to the time of first observation/2 is used.
    Disregards what is currently in the model.
    """
    mat_symb = _add_parameter(
        model, parameter_name, init=_get_absorption_init(model, parameter_name)
    )
    new_dose = Infusion(amount, duration=mat_symb * 2)
    to_comp.dose = new_dose


def _add_first_order_absorption(model, dose, to_comp):
    """Add first order absorption
    Disregards what is currently in the model.
    """
    odes = model.statements.ode_system
    depot = odes.add_compartment('DEPOT')
    depot.dose = dose
    mat_symb = _add_parameter(model, 'MAT', _get_absorption_init(model, 'MAT'))
    odes.add_flow(depot, to_comp, 1 / mat_symb)
    return depot


def _get_absorption_init(model, param_name):
    try:
        if param_name == 'MDT':
            param_prev = model.statements.lag_time
        else:
            param_prev = _extract_params_from_symb(model.statements, param_name, model.parameters)
        return param_prev.init
    except (AttributeError, KeyError):
        pass

    dt = model.dataset
    time_label = dt.pharmpy.idv_label
    obs = dt.pharmpy.observations
    time_min = obs.index.get_level_values(level=time_label).min()

    if param_name == 'MDT':
        return float(time_min) / 2
    elif param_name == 'MAT':
        return float(time_min) * 2


def set_peripheral_compartments(model, n):
    """Sets the number of peripheral compartments to a specified number.

    Parameters
    ----------
    model : Model
        Pharmpy model
    n : int
        Number of transit compartments

    Return
    ------
    Model
        Reference to same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> set_peripheral_compartments(model, 2)     # doctest: +ELLIPSIS
    <...>
    >>> model.statements.ode_system
    Bolus(AMT)
    ┌───────────┐
    │PERIPHERAL1│
    └───────────┘
      ↑      │
    QP1/V QP1/VP1
      │      ↓
    ┌───────────┐       ┌──────┐
    │  CENTRAL  │──CL/V→│OUTPUT│
    └───────────┘       └──────┘
       ↑      │
    QP2/VP2 QP2/V
       │      ↓
    ┌───────────┐
    │PERIPHERAL2│
    └───────────┘

    See also
    --------
    add_peripheral_compartment
    remove_peripheral_compartment

    """
    try:
        n = _as_integer(n)
    except TypeError:
        raise TypeError(f'Number of compartments must be integer: {n}')

    per = len(model.statements.ode_system.peripheral_compartments)
    if per < n:
        for _ in range(n - per):
            add_peripheral_compartment(model)
    elif per > n:
        for _ in range(per - n):
            remove_peripheral_compartment(model)
    return model


def add_peripheral_compartment(model):
    r"""Add a peripheral distribution compartment to model

    The rate of flow from the central to the peripheral compartment
    will be parameterized as QPn / VC where VC is the volume of the central compartment.
    The rate of flow from the peripheral to the central compartment
    will be parameterized as QPn / VPn where VPn is the volumne of the added peripheral
    compartment.

    Initial estimates:

    ==  ===================================================
    n
    ==  ===================================================
    1   :math:`\mathsf{CL} = \mathsf{CL'}`, :math:`\mathsf{VC} = \mathsf{VC'}`,
        :math:`\mathsf{QP1} = \mathsf{CL'}` and :math:`\mathsf{VP1} = \mathsf{VC'} \cdot 0.05`
    2   :math:`\mathsf{QP1} = \mathsf{QP1' / 2}`, :math:`\mathsf{VP1} = \mathsf{VP1'}`,
        :math:`\mathsf{QP2} = \mathsf{QP1' / 2}` and :math:`\mathsf{VP2} = \mathsf{VP1'}`
    ==  ===================================================

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Reference to same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> add_peripheral_compartment(model)     # doctest: +ELLIPSIS
    <...>
    >>> model.statements.ode_system
    Bolus(AMT)
    ┌───────────┐
    │PERIPHERAL1│
    └───────────┘
      ↑      │
    QP1/V QP1/VP1
      │      ↓
    ┌───────────┐       ┌──────┐
    │  CENTRAL  │──CL/V→│OUTPUT│
    └───────────┘       └──────┘

    See also
    --------
    set_peripheral_compartment
    remove_peripheral_compartment

    """
    statements = model.statements
    odes = statements.ode_system
    per = odes.peripheral_compartments
    n = len(per) + 1

    central = odes.central_compartment
    output = odes.output_compartment
    elimination_rate = odes.get_flow(central, output)
    cl, vc = elimination_rate.as_numer_denom()
    if cl.is_Symbol and vc == 1:
        # If K = CL / V
        s = statements.find_assignment(cl.name)
        cl, vc = s.expression.as_numer_denom()

    if n == 1:
        if vc == 1:
            kpc = _add_parameter(model, f'KPC{n}', init=0.1)
            kcp = _add_parameter(model, f'KCP{n}', init=0.1)
            peripheral = odes.add_compartment(f'PERIPHERAL{n}')
            odes.add_flow(central, peripheral, kcp)
            odes.add_flow(peripheral, central, kpc)
        else:
            full_cl = statements.before_odes.full_expression(cl)
            full_vc = statements.before_odes.full_expression(vc)
            pop_cl_candidates = full_cl.free_symbols & set(model.parameters.symbols)
            pop_cl = pop_cl_candidates.pop()
            pop_vc_candidates = full_vc.free_symbols & set(model.parameters.symbols)
            pop_vc = pop_vc_candidates.pop()
            pop_cl_init = model.parameters[pop_cl].init
            pop_vc_init = model.parameters[pop_vc].init
            qp_init = pop_cl_init
            vp_init = pop_vc_init * 0.05
    elif n == 2:
        per1 = per[0]
        from_rate = odes.get_flow(per1, central)
        qp1, vp1 = from_rate.as_numer_denom()
        full_qp1 = statements.before_odes.full_expression(qp1)
        full_vp1 = statements.before_odes.full_expression(vp1)
        pop_qp1_candidates = full_qp1.free_symbols & set(model.parameters.symbols)
        pop_qp1 = pop_qp1_candidates.pop()
        pop_vp1_candidates = full_vp1.free_symbols & set(model.parameters.symbols)
        pop_vp1 = pop_vp1_candidates.pop()
        pop_qp1_init = model.parameters[pop_qp1].init
        pop_vp1_init = model.parameters[pop_vp1].init
        model.parameters[pop_qp1].init = pop_qp1_init / 2
        qp_init = pop_qp1_init / 2
        vp_init = pop_vp1_init
    else:
        qp_init = 0.1
        vp_init = 0.1

    if vc != 1:
        qp = _add_parameter(model, f'QP{n}', init=qp_init)
        vp = _add_parameter(model, f'VP{n}', init=vp_init)
        peripheral = odes.add_compartment(f'PERIPHERAL{n}')
        odes.add_flow(central, peripheral, qp / vc)
        odes.add_flow(peripheral, central, qp / vp)

    return model


def remove_peripheral_compartment(model):
    r"""Remove a peripheral distribution compartment from model

    Initial estimates:

    ==  ===================================================
    n
    ==  ===================================================
    2   :math:`\mathsf{CL} = \mathsf{CL'}`,
        :math:`\mathsf{QP1} = \mathsf{CL'}` and :math:`\mathsf{VP1} = \mathsf{VC'} \cdot 0.05`
    3   :math:`\mathsf{QP1} = (\mathsf{QP1'} + \mathsf{QP2'}) / 2`,
        :math:`\mathsf{VP1} = \mathsf{VP1'} + \mathsf{VP2'}`
    ==  ===================================================

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Reference to same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> set_peripheral_compartments(model, 2)     # doctest: +ELLIPSIS
    <...>
    >>> remove_peripheral_compartment(model)      # doctest: +ELLIPSIS
    <...>
    >>> model.statements.ode_system
    Bolus(AMT)
    ┌───────────┐
    │PERIPHERAL1│
    └───────────┘
      ↑      │
    QP1/V QP1/VP1
      │      ↓
    ┌───────────┐       ┌──────┐
    │  CENTRAL  │──CL/V→│OUTPUT│
    └───────────┘       └──────┘

    See also
    --------
    set_peripheral_compartment
    add_peripheral_compartment

    """
    statements = model.statements
    odes = statements.ode_system
    peripherals = odes.peripheral_compartments
    if peripherals:
        last_peripheral = peripherals[-1]
        central = odes.central_compartment
        if len(peripherals) == 1:
            output = odes.output_compartment
            elimination_rate = odes.get_flow(central, output)
            cl, vc = elimination_rate.as_numer_denom()
            from_rate = odes.get_flow(last_peripheral, central)
            qp1, vp1 = from_rate.as_numer_denom()
            full_cl = statements.before_odes.full_expression(cl)
            full_vc = statements.before_odes.full_expression(vc)
            full_qp1 = statements.before_odes.full_expression(qp1)
            full_vp1 = statements.before_odes.full_expression(vp1)
            pop_cl_candidates = full_cl.free_symbols & set(model.parameters.symbols)
            pop_cl = pop_cl_candidates.pop()
            pop_vc_candidates = full_vc.free_symbols & set(model.parameters.symbols)
            pop_vc = pop_vc_candidates.pop()
            pop_qp1_candidates = full_qp1.free_symbols & set(model.parameters.symbols)
            pop_qp1 = pop_qp1_candidates.pop()
            pop_vp1_candidates = full_vp1.free_symbols & set(model.parameters.symbols)
            pop_vp1 = pop_vp1_candidates.pop()
            pop_vc_init = model.parameters[pop_vc].init
            pop_cl_init = model.parameters[pop_cl].init
            pop_qp1_init = model.parameters[pop_qp1].init
            pop_vp1_init = model.parameters[pop_vp1].init
            new_vc_init = pop_vc_init + pop_qp1_init / pop_cl_init * pop_vp1_init
            model.parameters[pop_vc].init = new_vc_init
        elif len(peripherals) == 2:
            first_peripheral = peripherals[0]
            from1_rate = odes.get_flow(first_peripheral, central)
            qp1, vp1 = from1_rate.as_numer_denom()
            from2_rate = odes.get_flow(last_peripheral, central)
            qp2, vp2 = from2_rate.as_numer_denom()
            full_qp2 = statements.before_odes.full_expression(qp2)
            full_vp2 = statements.before_odes.full_expression(vp2)
            full_qp1 = statements.before_odes.full_expression(qp1)
            full_vp1 = statements.before_odes.full_expression(vp1)
            pop_qp2_candidates = full_qp2.free_symbols & set(model.parameters.symbols)
            pop_qp2 = pop_qp2_candidates.pop()
            pop_vp2_candidates = full_vp2.free_symbols & set(model.parameters.symbols)
            pop_vp2 = pop_vp2_candidates.pop()
            pop_qp1_candidates = full_qp1.free_symbols & set(model.parameters.symbols)
            pop_qp1 = pop_qp1_candidates.pop()
            pop_vp1_candidates = full_vp1.free_symbols & set(model.parameters.symbols)
            pop_vp1 = pop_vp1_candidates.pop()
            pop_qp2_init = model.parameters[pop_qp2].init
            pop_vp2_init = model.parameters[pop_vp2].init
            pop_qp1_init = model.parameters[pop_qp1].init
            pop_vp1_init = model.parameters[pop_vp1].init
            new_qp1_init = (pop_qp1_init + pop_qp2_init) / 2
            new_vp1_init = pop_vp1_init + pop_vp2_init
            model.parameters[pop_qp1].init = new_qp1_init
            model.parameters[pop_vp1].init = new_vp1_init

        symbols = odes.get_flow(central, last_peripheral).free_symbols
        symbols |= odes.get_flow(last_peripheral, central).free_symbols
        odes.remove_compartment(last_peripheral)
        model.statements.remove_symbol_definitions(symbols, odes)
        model.remove_unused_parameters_and_rvs()
    return model


def set_ode_solver(model, solver):
    """Sets ODE solver to use for model

    Recognized solvers and their corresponding NONMEM advans:

    +------------------------+------------------+
    | Solver                 | NONMEM ADVAN     |
    +========================+==================+
    | CVODES                 | ADVAN14          |
    +------------------------+------------------+
    | DGEAR                  | ADVAN8           |
    +------------------------+------------------+
    | DVERK                  | ADVAN6           |
    +------------------------+------------------+
    | IDA                    | ADVAN15          |
    +------------------------+------------------+
    | LSODA                  | ADVAN13          |
    +------------------------+------------------+
    | LSODI                  | ADVAN9           |
    +------------------------+------------------+

    Parameters
    ----------
    model : Model
        Pharmpy model
    solver : str
        Solver to use or None for no preference

    Return
    ------
    Model
        Reference to same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> set_ode_solver(model, 'LSODA')    # doctest: +ELLIPSIS
    <...>

    """
    odes = model.statements.ode_system
    odes.solver = solver
    return model



from pharmpy.tools.rankfuncs import aic, bic, ofv


class DummyResults:
    def __init__(self, ofv=None, aic=None, bic=None, parameter_estimates=None):
        self.ofv = ofv
        self.aic = aic
        self.bic = bic
        self.parameter_estimates = parameter_estimates

    def __bool__(self):
        return bool(self.ofv) and bool(self.parameter_estimates)


class DummyModel:
    def __init__(self, name, **kwargs):
        self.name = name
        self.modelfit_results = DummyResults(**kwargs)


def test_ofv():
    run1 = DummyModel("run1", ofv=0)
    run2 = DummyModel("run2", ofv=-1)
    run3 = DummyModel("run3", ofv=-14)
    res = ofv(run1, [run2, run3])
    assert [run3] == res

    run4 = DummyModel("run4", ofv=2)
    run5 = DummyModel("run5", ofv=-2)
    res = ofv(run1, [run2, run3, run4, run5], cutoff=2)
    assert [run3, run5] == res


def test_aic():
    run1 = DummyModel("run1", aic=0)
    run2 = DummyModel("run2", aic=-1)
    run3 = DummyModel("run3", aic=-14)
    res = aic(run1, [run2, run3])
    assert [run3] == res


def test_bic():
    run1 = DummyModel("run1", bic=0)
    run2 = DummyModel("run2", bic=-1)
    run3 = DummyModel("run3", bic=-14)
    res = bic(run1, [run2, run3])
    assert [run3] == res



import sympy

import pharmpy.model
from pharmpy import Assignment, ModelStatements
from pharmpy.estimation import EstimationMethod
from pharmpy.workflows import Task, Workflow


def create_workflow(model=None):
    wf = Workflow()
    wf.name = "linearize"

    if model is not None:
        start_task = Task('start_linearize', start_linearize, model)
    else:
        start_task = Task('start_linearize', start_linearize)

    wf.add_task(start_task)
    return wf


def start_linearize(model):
    return model


def create_linearized_model(model):
    linbase = pharmpy.model.Model()
    linbase.parameters = model.parameters.copy()
    linbase.random_variables = model.random_variables.copy()

    ms = ModelStatements()
    base_terms_sum = 0
    for i, eta in enumerate(model.random_variables.etas, start=1):
        deta = sympy.Symbol("D_ETA1")
        oeta = sympy.Symbol("OETA")
        base = Assignment(f'BASE{i}', deta * (eta.symbol - oeta))
        ms.append(base)
        base_terms_sum += base.symbol

    base_terms = Assignment('BASE_TERMS', base_terms_sum)
    ms.append(base_terms)
    ipred = Assignment('IPRED', sympy.Symbol('OPRED') + base_terms.symbol)
    ms.append(ipred)

    i = 1
    err_terms_sum = 0
    for epsno, eps in enumerate(model.random_variables.epsilons, start=1):
        err = Assignment(f'ERR{epsno}', f'D_EPS{epsno}')
        err_terms_sum += err.symbol
        ms.append(err)
        i += 1
        for etano, eta in enumerate(model.random_variables.etas, start=1):
            inter = Assignment(
                f'ERR{i}',
                sympy.Symbol(f'D_EPSETA{epsno}_{etano}')
                * (eta.symbol - sympy.Symbol(f'OETA{etano}')),
            )
            err_terms_sum += inter.symbol
            ms.append(inter)
            i += 1
    error_terms = Assignment('ERROR_TERMS', err_terms_sum)
    ms.append(error_terms)

    y = model.dependent_variable
    Assignment(y, ipred.symbol + error_terms.symbol)
    linbase.statements = ms

    linbase.name = 'linbase'

    est = EstimationMethod('foce', interaction=True)
    linbase.estimation_steps = [est]
    return linbase



import copy
import uuid

import networkx as nx


class Workflow:
    def __init__(self, tasks=None):
        self._g = nx.DiGraph()
        if tasks:
            for task in tasks:
                self.add_task(task)

    def add_task(self, task, predecessors=None):
        """Add a task to the workflow.

        Predecessors will be connected if given.
        """
        self._g.add_node(task)
        if predecessors is not None:
            if not isinstance(predecessors, list):
                self._g.add_edge(predecessors, task)
            else:
                for pred in predecessors:
                    self._g.add_edge(pred, task)

    def insert_workflow(self, other, predecessors=None):
        if predecessors is None:
            output_tasks = self.output_tasks
        else:
            if isinstance(predecessors, list):
                output_tasks = predecessors
            else:
                output_tasks = [predecessors]
        input_tasks = other.input_tasks
        self._g = nx.compose(self._g, other._g)
        if len(input_tasks) == len(output_tasks):
            for inp, outp in zip(input_tasks, output_tasks):
                self._g.add_edge(outp, inp)
        elif len(input_tasks) == 1:
            for outp in output_tasks:
                self._g.add_edge(outp, input_tasks[0])
        elif len(output_tasks) == 1:
            for inp in input_tasks:
                self._g.add_edge(output_tasks[0], inp)
        else:
            raise ValueError('Having N:M connections between workflows is currently not supported')

    def as_dask_dict(self):
        ids = dict()
        for task in self._g.nodes():
            ids[task] = f'{task.name}-{uuid.uuid4()}'

        if len(self.output_tasks) == 1:
            ids[self.output_tasks[0]] = 'results'
        else:
            raise ValueError("Workflow can only have one output task")

        as_dict = dict()
        for task in nx.dfs_tree(self._g):
            key = ids[task]
            if task.has_input():
                input_list = list(task.task_input)
            else:
                input_list = []
            input_list.extend([ids[t] for t in self._g.predecessors(task)])
            value = (task.function, *input_list)
            as_dict[key] = value
        return as_dict

    @property
    def input_tasks(self):
        return [node for node in self._g.nodes if self._g.in_degree(node) == 0]

    @property
    def output_tasks(self):
        return [node for node in self._g.nodes if self._g.out_degree(node) == 0]

    def get_upstream_tasks(self, task):
        edges = list(nx.edge_dfs(self._g, task, orientation='reverse'))
        return [node for node, _, _ in edges]

    def get_predecessors(self, task):
        return list(self._g.predecessors(task))

    def get_successors(self, task):
        return list(self._g.successors(task))

    def copy(self):
        wf_copy = copy.deepcopy(self)
        return wf_copy

    def plot_dask(self, filename):
        from dask import visualize

        visualize(self.as_dask_dict(), filename=filename, collapse_outputs=True)

    @property
    def tasks(self):
        """All tasks"""
        return list(self._g.nodes())

    def __len__(self):
        return len(self._g.nodes)

    def __str__(self):
        return '\n'.join([str(task) for task in self._g])


class Task:
    def __init__(self, name, function, *task_input):
        self.name = name
        self.function = function
        self.task_input = task_input

    def has_input(self):
        return len(self.task_input) > 0

    def __repr__(self):
        return self.name



"""
Generic parser using lark-parser.

Subclass :class:`GenericParser` (remember to set :attr:`GenericParser.grammar`
to point to your grammar file) to define a powerful parser.
"""
import copy
import re

from lark import Tree
from lark.lexer import Token

from . import prettyprint


class NoSuchRuleException(AttributeError):
    """Rule not found (raised by :class:`AttrTree` for unknown children)."""

    def __init__(self, rule, tree=None):
        try:
            post = ' (%s)' % (repr(tree.rule),)
        except AttributeError:
            post = ''
        super().__init__('no %s child in tree%s' % (repr(rule), post))


def rule_of(item):
    """Rule of a tree or token. Convenience function (will not raise)."""

    try:
        return item.data
    except AttributeError:
        try:
            return item.type
        except AttributeError:
            return ''


def empty_rule(rule):
    """Create empty Tree or Token, depending on (all) upper-case or not."""

    if rule == rule.upper():
        return Token(rule, '')
    return Tree(rule, [])


class AttrToken(Token):
    """Token with attribute access.

    Created by :meth:`AttrTree.transform` from :class:`lark.lexer.Token`.

    Attributes:
        self.rule: Name, in common with :class:`AttrTree`.
        self.eval: Transformed data type, in common with :class:`AttrTree`.

    Can be instantiated with :meth:`.__init__`(rule, content), via :class:`lark.lexer.Token`, or
    alternative constructor :meth:`.transform` (transform object of class
    :class:`lark.lexer.Token`).
    """

    __slots__ = ()

    @classmethod
    def transform(cls, token, **kwargs):
        """Alternative constructor: From Token (with optional overrides)."""
        kwargs = {
            'type_': token.type,
            'value': token.value,
            'start_pos': token.start_pos,
            'line': token.line,
            'column': token.column,
            **kwargs,
        }
        return cls(**kwargs)

    @property
    def rule(self):
        """Rule name (synonymous with 'type')"""
        return self.type

    @rule.setter
    def rule(self, value):
        self.type = value

    @property
    def eval(self):
        """Evaluated value (str, int, float)."""
        if self.type in {'DIGIT', 'INT', 'SIGNED_INT'}:
            return int(self.value)
        elif self.type in {
            'DECIMAL',
            'EXP',
            'FLOAT',
            'NUMBER',
            'NUMERIC',
            'SIGNED_FLOAT',
            'SIGNED_NUMBER',
        }:
            return float(self.value)
        elif self.type == 'NEG_INF':
            return float('-INF')
        elif self.type == 'POS_INF':
            return float('INF')
        else:
            return str(self.value)

    def replace(self, value):
        """Returns copy (same rule), but with content replaced."""
        return self.transform(token=self, value=value)

    def __deepcopy__(self, memo):
        return AttrToken(self.type, self.value, self.start_pos, self.line, self.column)

    def __repr__(self):
        return '%s(%s, %s)' % (self.__class__.__name__, repr(self.rule), repr(self.value))

    def __hash__(self):
        return hash(repr(self))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __str__(self):
        return str(self.value)


class AttrTree(Tree):
    """Tree with attribute access.

    Created in :meth:`GenericParser.parse` by :meth:`transform`, from :class:`lark.Tree`.

    Attributes:
        self.rule: Name, in common with :class:`.AttrToken`.
        self.rules: Names of children.
        self.eval: Transformed data type, in common with :class:`.AttrToken`.
        self.tokens: Recursive tokens as (flattened) list.
        self.debug: Treeview str, formatted for debugging.

    Can be instantiated with :meth:`.__init__`, via :class:`lark.Tree`, or alternative constructors:
        1. :meth:`.transform` (transform recursively object of class :class:`lark.Tree`).
        2. :meth:`.create` (create from nested iterators).
    """

    AttrToken = AttrToken

    @classmethod
    def transform(cls, tree, **kwargs):
        """Alternative constructor: From Tree (with optional overrides)."""
        kwargs = {'data': tree.data, 'children': tree.children, 'meta': tree._meta, **kwargs}
        children = kwargs['children'].copy()
        for i, child in enumerate(children):
            if isinstance(child, Tree):
                children[i] = cls.transform(tree=child)
            elif isinstance(child, Token):
                children[i] = cls.AttrToken.transform(token=child)
            else:
                children[i] = cls.AttrToken('STRING', str(child))
        kwargs['children'] = children
        return cls(**kwargs)

    @classmethod
    def create(cls, rule, items, _anon_count=0, _list=False):
        """Alternative constructor: Creates new tree from (possibly nested) iterables.

        Only non-iterable items become leaves (i.e. content of token nodes). All others are trees.

        Handling of missing names (e.g. lists):
            1. Tree: Children are moved to parent.
            2. Token: Native naming scheme, __ANON_%d, of Lark is used.

        Args:
            items: Child (tree) nodes or content (token) nodes.
            rule: Name of root tree (__ANON_0 if False).
            _anon_count: Internal. Anonymous numbering offset.
            _list: Internal. Recursion state. Drop 'rule' & return list of children, which are
                orphaned if name is False.

        Raises:
            TypeError: 'items' not iterable or instance of 'str' (only leaves shall contain 'str').
            ValueError: 'items' empty (trees can't be empty).

        .. note:: Please follow convention of all lower/upper case for trees/tokens.
        """

        def non_iterable(rule, items):
            raise TypeError(
                '%s object is not iterable (of children for tree %s)'
                % repr(items.__class__.__name__),
                repr(rule),
            )

        # determine mode of operation; 'items' dict-like OR just iterable?
        try:
            names, items = zip(*items.items())
        except AttributeError:
            if isinstance(items, str):
                non_iterable(rule, items)
            try:
                length = len(items)
            except TypeError:
                non_iterable(rule, items)
            names = [None] * length
        if not items:
            raise ValueError('refusing empty tree %s (only tokens are childless)' % repr(rule))

        # create the nodes
        new_nodes = []
        for name, thing in zip(names, items):
            try:  # try to recurse down
                nodes, _anon_count = cls.create(name, thing, _anon_count=_anon_count, _list=True)
            except TypeError:  # looks like a leaf
                try:  # don't convert existing nodes (to leaves)
                    name = thing.rule
                except AttributeError:
                    if not name:
                        _anon_count += 1
                        name = '__ANON_%d' % (_anon_count,)
                    new_nodes += [cls.AttrToken(name, str(thing))]
                else:  # node already, won't recreate
                    new_nodes += [thing]
            else:
                new_nodes += nodes
        # list (and counter) for recursion
        if _list:
            return [cls(rule, new_nodes)] if rule else new_nodes, _anon_count

        # tree to external caller
        if rule:
            return cls(rule, new_nodes)
        return cls('__ANON_0', new_nodes)

    # -- public interface ----------------------------------------------
    @property
    def rule(self):
        """Rule name (synonymous with 'data')."""
        return self.data

    @rule.setter
    def rule(self, value):
        self.data = value

    @property
    def rules(self):
        """All rules of (immediate) children."""
        return [node.rule for node in self.children]

    @property
    def eval(self):
        """Evaluated value (self)."""
        return self

    def find(self, rule):
        """Returns first child matching 'rule', or None."""
        return next((child for child in self.children if child.rule == rule), None)

    def all(self, rule):
        """Returns all children matching rule, or []."""
        return list(filter(lambda child: child.rule == rule, self.children))

    def set(self, rule, value):
        """Sets first child matching rule. Raises if none."""
        for i, child in enumerate(self.children):
            if child.rule == rule:
                self.children[i] = value
                return
        raise NoSuchRuleException(rule, self)

    def partition(self, rule):
        """Partition children into (head, item, tail).

        Search for child item 'rule' and return the part before it (head), the item, and the part
        after it (tail). If 'rule' is not found, return (children, [], [])."""
        head, item, tail = [], [], []
        for node in self.children:
            if not item and node.rule == rule:
                item += [node]
            elif not item:
                head += [node]
            else:
                tail += [node]
        return (head, item, tail)

    def tree_walk(self):
        """Generator for iterating depth-first (i.e. parse order) over children."""
        for child in self.children:
            yield child
            try:
                yield from child.tree_walk()
            except AttributeError:
                continue

    def remove(self, rule):
        """Remove all children with rule. Not recursively"""
        new_children = []
        for child in self.children:
            if child.rule != rule:
                new_children.append(child)
        self.children = new_children

    def remove_node(self, node):
        new_children = []
        comment_flag = False
        for child in self.children:
            if child.rule == 'COMMENT' or child.rule == 'NEWLINE':
                if not comment_flag:
                    new_children.append(child)
                else:
                    if str(child).startswith('\n\n'):
                        new_children.append(child.replace(str(child)[1:]))
                    comment_flag = False
            elif str(child.eval) != str(node.eval):
                new_children.append(child)
                if child.rule == 'statement':
                    comment_flag = False
            else:
                comment_flag = True
        self.children = new_children

    @staticmethod
    def _newline_node():
        return AttrToken('WS_ALL', '\n')

    def _clean_ws(self, new_children):
        new_children_clean = []
        types_of_newline = ['WS_ALL', 'NEWLINE']
        last_index = len(new_children) - 1

        prev_rule = None

        for i, child in enumerate(new_children):
            if child.rule in types_of_newline:
                if prev_rule in types_of_newline or prev_rule == 'verbatim':
                    continue
                if i == last_index:
                    continue
            if re.search('\n{2,}', str(child)):
                new_children_clean.append(self._newline_node())
            else:
                new_children_clean.append(child)

            prev_rule = child.rule

        return new_children_clean

    def add_node(self, node, following_node=None, comment=False):
        new_children = copy.deepcopy(self.children)

        if comment:
            new_children = self._clean_ws(new_children)

        if following_node is None:
            if not comment:
                new_children.append(self._newline_node())
            new_children.append(node)
        else:
            index = self.children.index(following_node)
            new_children.insert(index, node)

            newline_node = copy.deepcopy(new_children[index - 1])
            new_children.insert(index + 1, newline_node)

        new_children_clean = self._clean_ws(new_children)
        self.children = new_children_clean

    def add_comment_node(self, comment, adjacent_node=None):
        comment_node = AttrToken('COMMENT', f' ; {comment}')
        self.add_node(comment_node, adjacent_node, comment=True)

    def add_newline_node(self):
        self.children.append(self._newline_node())

    def get_last_node(self):
        for node in self.children:
            last_node = node
        return last_node

    @property
    def tokens(self):
        """All tokens as flattened list."""
        items = []
        for item in self.children:
            try:
                items += item.tokens
            except AttributeError:
                items += [item]
        return items

    @property
    def debug(self, *args, **kwargs):
        """Debug formatted tree structure."""
        return str(prettyprint.transform(self, *args, **kwargs))

    def treeprint(self, indent=''):
        """Prints debug formatted tree structure."""
        print(self.debug)

    def set_child(self, attr, value):
        self.find(attr).value = value

    # -- private methods -----------------------------------------------
    def __len__(self):
        return len(self.children)

    def __setattr__(self, attr, value):
        if attr in ['data', 'children', '_meta']:
            object.__setattr__(self, attr, value)
        else:
            self.set_child(attr, value)

    def __getattr__(self, attr):
        if attr in ['data', 'children', '_meta']:
            return object.__getattribute__(self, attr)
        child = self.find(attr)
        if child is None:
            raise NoSuchRuleException(attr, self)
        return child.eval

    def __str__(self):
        return ''.join(str(x) for x in self.children)

    def __repr__(self):
        return '%s(%s, %s)' % (self.__class__.__name__, repr(self.rule), repr(self.children))

    def __hash__(self):
        return hash(repr(self))

    def __eq__(self, other):
        return hash(self) == hash(other)


class GenericParser:
    """
    Generic parser using lark-parser.

    Inherit to define a parser, say ThetaRecordParser for NONMEM, with the workflow:

    1. Lex and parse a 'buffer' using Lark_ (from :attr:`grammar` file). Builds AST.
    2. Shape AST by options (see :attr:`non_empty`).
    3. Convert to :class:`AttrTree` for convenient traversal (attribute access).

    Attributes:
        self.non_empty: Insert empty placeholders if missing. Dict of rule -> (pos, name), where a
            Tree or Token (if uppercase) will be inserted at 'pos' of the children of 'rule', if
            none exists.
        self.buffer: Buffer parsed by :meth:`parse`.
        self.grammar: Path to grammar file.
        self.root: Root of final tree. Instance of :class:`AttrTree`.

    .. _Lark:
        https://github.com/lark-parser/lark
    """

    """Children to create if missing"""
    non_empty = []

    """:class:`AttrTree` implementation."""
    AttrTree = AttrTree

    lark_options = dict(
        ambiguity='resolve',
        debug=False,
        keep_all_tokens=True,
        lexer='dynamic',
        parser='earley',
        start='root',
    )

    def __init__(self, buf=None, **lark_options):
        self.lark_options.update(lark_options)
        self.root = self.parse(buf)

    def parse(self, buf):
        """Parses a buffer, transforms and constructs :class:`AttrTree` object.

        Args:
            buf: Buffer to parse.
        """
        self.buffer = buf
        if self.buffer is None:
            return None

        root = self.lark.parse(self.buffer)
        if self.non_empty:
            root = self.insert(root, self.non_empty)
        return self.AttrTree.transform(tree=root)

    @classmethod
    def insert(cls, item, non_empty):
        """Inserts missing Tree/Token amongst children (see :attr:`non_empty`).

        Args:
            item: Tree to recurse.
            non_empty: Dict of rule -> (pos, name) tuple.
        """
        if not non_empty or isinstance(item, Token):
            return item
        for d in non_empty:
            try:
                pos, name = d[rule_of(item)]
            except KeyError:
                pass
            else:
                if not any(rule_of(child) == name for child in item.children):
                    item.children.insert(pos, empty_rule(name))
        for i, child in enumerate(item.children):
            item.children[i] = cls.insert(child, non_empty)
        return item

    def __str__(self):
        if not self.root:
            return repr(self)
        lines = str(prettyprint.transform(self.root)).splitlines()
        return '\n'.join(lines)


def remove_token_and_space(tree, rule, recursive=False):
    """Remove all tokens with rule and any WS before it"""
    new_nodes = []
    for node in tree.children:
        if node.rule == rule:
            if len(new_nodes) > 0 and new_nodes[-1].rule == 'WS':
                new_nodes.pop()
        else:
            new_nodes.append(node)
    tree.children = new_nodes
    if recursive:
        for node in tree.children:
            if hasattr(node, 'children'):
                remove_token_and_space(node, rule, recursive=True)


def insert_before_or_at_end(tree, rule, nodes):
    """Insert nodes before rule or if rule does not exist at end"""
    kept = []
    found = False
    for node in tree.children:
        if node.rule == rule:
            kept.extend(nodes)
            found = True
        kept.append(node)
    if not found:
        kept.extend(nodes)
    tree.children = kept


def insert_after(tree, rule, nodes):
    """Insert nodes after rule"""
    kept = []
    for node in tree.children:
        kept.append(node)
        if node.rule == rule:
            kept.extend(nodes)
    tree.children = kept



import pytest

from pharmpy.workflows import Task, Workflow
from pharmpy.workflows.databases.local_directory import LocalDirectoryDatabase
from pharmpy.workflows.dispatcher import ExecutionDispatcher
from pharmpy.workflows.dispatchers.local import LocalDispatcher


def fun(s):
    return s


def test_base_class():
    disp = ExecutionDispatcher()
    with pytest.raises(NotImplementedError):
        disp.run(None, None)


def test_local_dispatcher():
    db = LocalDirectoryDatabase()
    disp = LocalDispatcher()
    wf = Workflow([Task('results', fun, 'input')])
    res = disp.run(wf, db)
    assert res == 'input'



# import logging

# import pytest


# @pytest.mark.asyncio
# async def test_pheno_execute(pheno, event_loop):
#    """Test actual execution of pheno."""
#    logger = logging.getLogger()

#    copy = pheno.copy()
#    rundir = await copy.execute.estimate()
#    logger.debug('%r._jobs=%r', rundir, rundir._jobs)

#    job = rundir._jobs[0]
#    assert not job.done
#    await job.wait(10, 0.5)
#    assert job.done
# assert job.rc == 0

#    path = rundir.path
#    assert path.exists()



import pytest

import pharmpy


@pytest.fixture(scope='session')
def datadir(testdata):
    return testdata / 'nonmem'


@pytest.fixture(scope='session')
def pheno_path(datadir):
    return datadir / 'pheno_real.mod'


@pytest.fixture(scope='session')
def pheno_ext(datadir):
    return datadir / 'pheno_real.ext'


@pytest.fixture(scope='session')
def pheno_phi(datadir):
    return datadir / 'pheno_real.phi'


@pytest.fixture(scope='session')
def pheno_cov(datadir):
    return datadir / 'pheno_real.cov'


@pytest.fixture(scope='session')
def pheno_lst(datadir):
    return datadir / 'pheno_real.lst'


@pytest.fixture(scope='session')
def pheno_data(datadir):
    return datadir / 'pheno.dta'


@pytest.fixture(scope='session')
def pheno(pheno_path):
    model = pharmpy.Model(pheno_path)
    return model



from pharmpy import Model


def test_dataset(testdata):
    model = Model(testdata / 'nonmem' / 'fcon' / 'FCON')
    assert model.code.startswith('FILE')

    df = model.dataset

    nmtran = Model(testdata / 'nonmem' / 'pheno_real.mod')

    assert list(df['WGT']) == list(nmtran.dataset['WGT'])



import itertools
import warnings
from pathlib import Path

from pharmpy import Model
from pharmpy.data import ColumnType
from pharmpy.data_structures import OrderedSet

from .models import create_model3b


def setup(model_path, covariates):
    input_model = Model(model_path)
    covariates = check_covariates(input_model, covariates)
    return covariates


def check_covariates(input_model, covariates):
    """Perform checks of covariates and filter out inappropriate ones

    1. A covariate has the same baseline value for all individuals
    2. Two or more covariates have the exact same base line values
    3. Warn for time varying covariates

    Return a new list of covariates with "bad" ones removed
    """
    data = input_model.dataset
    data.pharmpy.column_type[covariates] = ColumnType.COVARIATE
    cov_bls = data.pharmpy.covariate_baselines

    tvar = data.pharmpy.time_varying_covariates
    if tvar:
        warnings.warn(
            f'The covariates {tvar} are time varying, but FREM will only use the '
            f'baseline values.'
        )

    # Covariates with only one baseline value
    unique = (cov_bls.iloc[0] == cov_bls.iloc[1:]).all()
    covset = OrderedSet(covariates)
    unique_covariates = OrderedSet(list((unique[unique].index)))
    covset -= unique_covariates
    if unique_covariates:
        warnings.warn(
            f'The covariates {list(unique_covariates)} have the same baseline value for '
            f'all individuals and has been removed from the analysis.'
        )

    # Remove covariates that are equal
    new_covset = OrderedSet(covset)
    for col1, col2 in itertools.combinations(covset, 2):
        if cov_bls[col1].equals(cov_bls[col2]):
            new_covset.discard(col2)
            warnings.warn(
                f'The baselines for covariates {col1} and {col2} are equal for all '
                f'individuals. {col2} has been removed from the analysis.'
            )

    return list(new_covset)


def update_model3b_for_psn(rundir, ncovs):
    """Function to update model3b from psn

    NOTE: This function lets pharmpy tie in to the PsN workflow
          and is a temporary solution
    """
    model_path = Path(rundir) / 'm1'
    model1b = Model(model_path / 'model_1b.mod')
    model3 = Model(model_path / 'model_3.mod')
    model3b = create_model3b(model1b, model3, int(ncovs))
    model3b.write(model_path, force=True)



import numpy as np
import pandas as pd

import pharmpy.tools.rankfuncs as rankfuncs
import pharmpy.workflows as workflows

from .psn_helpers import tool_name


class Tool:
    def __init__(self, dispatcher=None, database=None, path=None):
        toolname = type(self).__name__.lower()
        if dispatcher is None:
            self.dispatcher = workflows.default_dispatcher
        else:
            self.dispatcher = dispatcher
        if database is None:
            self.database = workflows.default_tool_database(toolname=toolname, path=path)
        else:
            self.database = database


def create_results(path, **kwargs):
    name = tool_name(path)
    # FIXME: Do something automatic here
    if name == 'qa':
        from pharmpy.tools.qa.results import psn_qa_results

        res = psn_qa_results(path, **kwargs)
    elif name == 'bootstrap':
        from pharmpy.tools.bootstrap.results import psn_bootstrap_results

        res = psn_bootstrap_results(path, **kwargs)
    elif name == 'cdd':
        from pharmpy.tools.cdd.results import psn_cdd_results

        res = psn_cdd_results(path, **kwargs)
    elif name == 'frem':
        from pharmpy.tools.frem.results import psn_frem_results

        res = psn_frem_results(path, **kwargs)
    elif name == 'linearize':
        from pharmpy.tools.linearize.results import psn_linearize_results

        res = psn_linearize_results(path, **kwargs)
    elif name == 'resmod':
        from pharmpy.tools.resmod.results import psn_resmod_results

        res = psn_resmod_results(path, **kwargs)

    elif name == 'scm':
        from pharmpy.tools.scm.results import psn_scm_results

        res = psn_scm_results(path, **kwargs)
    elif name == 'simeval':
        from pharmpy.tools.simeval.results import psn_simeval_results

        res = psn_simeval_results(path, **kwargs)
    elif name == 'crossval':
        from pharmpy.tools.crossval.results import psn_crossval_results

        res = psn_crossval_results(path, **kwargs)
    else:
        raise ValueError("Not a valid run directory")
    return res


def create_summary(models, start_model, rankfunc, cutoff, model_features):
    rankfunc = getattr(rankfuncs, rankfunc)

    res_data = {'dofv': [], 'features': [], 'rank': []}
    model_names = []

    if cutoff is not None:
        ranks = rankfunc(start_model, models, cutoff=cutoff)
    else:
        ranks = rankfunc(start_model, models)

    for model in models:
        model_names.append(model.name)
        res_data['dofv'].append(start_model.modelfit_results.ofv - model.modelfit_results.ofv)
        res_data['features'].append(model_features[model.name])
        if model in ranks:
            res_data['rank'].append(ranks.index(model) + 1)
        else:
            res_data['rank'].append(np.nan)

    # FIXME: in ranks, if any row has NaN the rank converts to float
    return pd.DataFrame(res_data, index=model_names)



import math

import numpy as np
import sympy

# This module could probably be made private.


def triangular_root(x):
    """Calculate the triangular root of x. I.e. if x is a triangular number T_n what is n?"""
    return math.floor(math.sqrt(2 * x))


def flattened_to_symmetric(x):
    """Convert a vector containing the elements of a lower triangular matrix into a full symmetric
    matrix
    """
    n = triangular_root(len(x))
    new = np.zeros((n, n))
    inds = np.tril_indices_from(new)
    new[inds] = x
    new[(inds[1], inds[0])] = x
    return new


def cov2corr(cov):
    """Convert covariance matrix to correlation matrix"""
    v = np.sqrt(np.diag(cov))
    outer_v = np.outer(v, v)
    corr = cov / outer_v
    corr[cov == 0] = 0
    return corr


def corr2cov(corr, sd):
    """Convert correlation matrix to covariance matrix

    Parameters
    ----------
    corr
        Correlation matrix (ones on diagonal)
    sd
        One dimensional array of standard deviations
    """
    sd_matrix = np.diag(sd)
    cov = sd_matrix @ corr @ sd_matrix
    return cov


def round_and_keep_sum(x, s):
    """Round values in Series x and their sum must be s

    Algorithm: Floor all elements in series. If sum not correct add one to element with
               highest fractional part until sum is reached.
    """
    sorted_fractions = x.apply(lambda x: math.modf(x)[0]).sort_values(ascending=False)
    rounded_sample_sizes = x.apply(lambda x: math.modf(x)[1])
    for (group_index, _) in sorted_fractions.iteritems():
        num_samples = rounded_sample_sizes.sum()
        diff = s - num_samples
        if diff == 0:
            break
        step = math.copysign(1, diff)
        rounded_sample_sizes[group_index] += step

    return rounded_sample_sizes.astype('int64')


def se_delta_method(expr, values, cov):
    """Use the delta method to estimate the standard error of a function of parameters with covariance
    matrix available.

    Parameters
    ----------
    expr
        A sympy expression for the function of parameters
    cov
        Dataframe with symbol names as indices must include at least all parameters needed for expr
    values
        dict/series parameter estimates. Must include at least all parameters needed for expr

    Returns
    -------
    Approximation of the standard error
    """
    symbs = expr.free_symbols
    names_unsorted = [s.name for s in symbs]
    # Sort names according to order in cov
    names = [y for x in cov.columns for y in names_unsorted if y == x]
    cov = cov[names].loc[names]
    symb_gradient = [sympy.diff(expr, sympy.Symbol(name)) for name in names]
    num_gradient = np.array([float(x.subs(values)) for x in symb_gradient])
    se = np.sqrt(num_gradient @ cov.values @ num_gradient.T)
    return se


def is_posdef(A):
    """Checks whether a matrix is positive definite"""
    try:
        _ = np.linalg.cholesky(A)
        return True
    except np.linalg.LinAlgError:
        return False


def nearest_posdef(A):
    """Return the nearest positive definite matrix in the Frobenius norm to a matrix

    A Python/Numpy port of John D'Errico's `nearestSPD` MATLAB code [1], which
    credits [2].
    [1] https://www.mathworks.com/matlabcentral/fileexchange/42885-nearestspd
    [2] N.J. Higham, "Computing a nearest symmetric positive semidefinite
    matrix" (1988): https://doi.org/10.1016/0024-3795(88)90223-6
    [3] https://gist.github.com/fasiha/fdb5cec2054e6f1c6ae35476045a0bbd
    """
    if is_posdef(A):
        return A

    B = (A + A.T) / 2
    _, s, V = np.linalg.svd(B)

    H = np.dot(V.T, np.dot(np.diag(s), V))
    A2 = (B + H) / 2
    A3 = (A2 + A2.T) / 2

    if is_posdef(A3):
        return A3

    spacing = np.spacing(np.linalg.norm(A))
    # The above is different from [1]. It appears that MATLAB's `chol` Cholesky
    # decomposition will accept matrixes with exactly 0-eigenvalue, whereas
    # Numpy's will not. So where [1] uses `eps(mineig)` (where `eps` is Matlab
    # for `np.spacing`), we use the above definition. CAVEAT: our `spacing`
    # will be much larger than [1]'s `eps(mineig)`, since `mineig` is usually on
    # the order of 1e-16, and `eps(1e-16)` is on the order of 1e-34, whereas
    # `spacing` will, for Gaussian random matrixes of small dimension, be on
    # othe order of 1e-16. In practice, both ways converge, as the unit test
    # below suggests.
    Id = np.eye(A.shape[0])
    k = 1
    while not is_posdef(A3):
        mineig = np.min(np.real(np.linalg.eigvals(A3)))
        A3 += Id * (-mineig * k ** 2 + spacing)
        k += 1

    return A3


def conditional_joint_normal(mu, sigma, a):
    """Give parameters of the conditional joint normal distribution

    The condition is the last len(a) values

    See https://en.wikipedia.org/wiki/Multivariate_normal_distribution#Conditional_distributions
    """

    # partition mu and sigma
    nfirst = len(mu) - len(a)
    S11 = sigma[0:nfirst, 0:nfirst]
    S12 = sigma[0:nfirst, nfirst:]
    S21 = sigma[nfirst:, 0:nfirst]
    S22 = sigma[nfirst:, nfirst:]
    M1 = mu[0:nfirst]
    M2 = mu[nfirst:]

    S22_inv = np.linalg.inv(S22)

    mu_bar = M1 + S12 @ S22_inv @ (a - M2)
    sigma_bar = S11 - S12 @ S22_inv @ S21

    return mu_bar, sigma_bar


def round_to_n_sigdig(x, n):
    if x == 0:
        return x
    else:
        return round(x, -int(math.floor(math.log10(abs(x)))) + (n - 1))


def is_near_target(x, target, zero_limit, significant_digits):
    if target == 0:
        return abs(x) < abs(zero_limit)
    else:
        return round_to_n_sigdig(x, n=significant_digits) == round_to_n_sigdig(
            target, n=significant_digits
        )



import pytest

from pharmpy.estimation import EstimationMethod


def test_estimation_method():
    a = EstimationMethod('foce', cov=True)
    assert a.method == 'FOCE'
    assert a.cov

    with pytest.raises(ValueError):
        EstimationMethod('sklarfs')

    a.method = 'fo'
    assert a.method == 'FO'

    assert a == EstimationMethod('fo', interaction=False, cov=True)
    assert (
        repr(a) == 'EstimationMethod("FO", interaction=False, cov=True, evaluation=False, '
        'maximum_evaluations=None, laplace=False, tool_options=None)'
    )

    with pytest.raises(ValueError):
        EstimationMethod('fo', maximum_evaluations=0)


def test_append_options():
    a = EstimationMethod('foce')

    a.append_tool_options({'EONLY': 1})
    assert len(a.tool_options) == 1

    a.append_tool_options({'PRINT': 1})
    assert len(a.tool_options) == 2


def test_copy(datadir):
    a = EstimationMethod('foce', cov=True)
    b = a.copy()
    assert id(a) != id(b)
    assert b.method == 'FOCE'



"""
The NONMEM $MODEL record
"""

from .option_record import OptionRecord


class ModelRecord(OptionRecord):
    def add_compartment(self, name, dosing=False):
        options = [name]
        if dosing:
            options.append('DEFDOSE')
        self.append_option('COMPARTMENT', f'({" ".join(options)})')

    def prepend_compartment(self, name, dosing=False):
        options = [name]
        if dosing:
            options.append('DEFDOSE')
        self.prepend_option('COMPARTMENT', f'({" ".join(options)})')

    def get_compartment_number(self, name):
        for i, (curname, _) in enumerate(self.compartments()):
            if name == curname:
                return i + 1
        return None

    def remove_compartment(self, name):
        n = self.get_compartment_number(name)
        self.remove_nth_option('COMPARTMENT', n - 1)

    def set_dosing(self, name):
        n = self.get_compartment_number(name)
        self.add_suboption_for_nth('COMPARTMENT', n - 1, 'DEFDOSE')

    def move_dosing_first(self):
        self.remove_suboption_for_all('COMPARTMENT', 'DEFDOSE')
        self.add_suboption_for_nth('COMPARTMENT', 0, 'DEFDOSE')

    def compartments(self):
        all_options = [
            'INITIALOFF',
            'NOOFF',
            'NODOSE',
            'EQUILIBRIUM',
            'EXCLUDE',
            'DEFOBSERVATION',
            'DEFDOSE',
        ]
        for n, opts in enumerate(self.get_option_lists('COMPARTMENT')):
            name = f'COMP{n + 1}'
            options = []
            for opt in opts:
                match = OptionRecord.match_option(all_options, opt)
                if match:
                    options.append(match)
                else:
                    name = opt
            yield name, options



"""
:meta private:
"""

import math
import warnings
from operator import add, mul

import numpy as np
import sympy
from sympy import Eq, Float, Gt, Le, Piecewise, exp

from pharmpy.parameter import Parameter
from pharmpy.statements import Assignment, ModelStatements
from pharmpy.symbols import symbol as S


def add_covariate_effect(model, parameter, covariate, effect, operation='*'):
    """Adds covariate effect to :class:`pharmpy.model`.

    The following effects have templates:

    - Linear function for continuous covariates (*lin*)
        - Function:

        .. math::

            \\text{coveff} = 1 + \\text{theta} * (\\text{cov} - \\text{median})

        - Init:  0.001
        - Upper:
            - If median of covariate equals minimum: :math:`100,000`
            - Otherwise: :math:`\\frac{1}{\\text{median} - \\text{min}}`
        - Lower:
            - If median of covariate equals maximum: :math:`-100,000`
            - Otherwise: :math:`\\frac{1}{\\text{median} - \\text{max}}`
    - Linear function for categorical covariates (*cat*)
        - Function:

            - If covariate is most common category:

            .. math::

                \\text{coveff} = 1

            - For each additional category:

            .. math::

                \\text{coveff} = 1 + \\text{theta}

        - Init: :math:`0.001`
        - Upper: :math:`100,000`
        - Lower: :math:`-100,000`
    - Piecewise linear function/"hockey-stick", continuous covariates only (*piece_lin*)
        - Function:
            - If cov <= median:

            .. math::

                \\text{coveff} = 1 + \\text{theta1} * (\\text{cov} - \\text{median})

            - If cov > median:

            .. math::

                \\text{coveff} = 1 + \\text{theta2} * (\\text{cov} - \\text{median})


        - Init: :math:`0.001`
        - Upper:
            - For first state: :math:`\\frac{1}{\\text{median} - \\text{min}}`
            - Otherwise: :math:`100,000`
        - Lower:
            - For first state: :math:`-100,000`
            - Otherwise: :math:`\\frac{1}{\\text{median} - \\text{max}}`
    - Exponential function, continuous covariates only (*exp*)
        - Function:

        .. math::

            \\text{coveff} = \\exp(\\text{theta} * (\\text{cov} - \\text{median}))

        - Init:
            - If lower > 0.001 or upper < 0.001: :math:`\\frac{\\text{upper} - \\text{lower}}{2}`
            - If estimated init is 0: :math:`\\frac{\\text{upper}}{2}`
            - Otherwise: :math:`0.001`
        - Upper:
            - If min - median = 0 or max - median = 0: :math:`100`
            - Otherwise:

            .. math::

                \\min(\\frac{\\log(0.01)}{\\text{min} - \\text{median}},
                \\frac{\\log(100)}{\\text{max} - \\text{median}})
        - Lower:
            - If min - median = 0 or max - median = 0: :math:`0.01`
            - Otherwise:

            .. math::

                \\max(\\frac{\\log(0.01)}{\\text{max} - \\text{median}},
                \\frac{\\log(100)}{\\text{min} - \\text{median}})

    - Power function, continuous covariates only (*pow*)
        - Function:

        .. math::

            \\text{coveff} = (\\frac{\\text{cov}}{\\text{median}})^\\text{theta}

        - Init: :math:`0.001`
        - Upper: :math:`100,000`
        - Lower: :math:`-100`


    Parameters
    ----------
    model : Model
        Pharmpy model to add covariate effect to.
    parameter : str
        Name of parameter to add covariate effect to.
    covariate : str
        Name of covariate.
    effect : str
        Type of covariate effect. May be abbreviated covariate effect (see above) or custom.
    operation : str, optional
        Whether the covariate effect should be added or multiplied (default).

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> add_covariate_effect(model, "CL", "APGR", "exp")    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.before_odes.full_expression("CL")
    THETA(1)*WGT*exp(ETA(1))*exp(POP_CLAPGR*(APGR - 7.0))

    """
    sset = model.statements

    if S(f'{parameter}{covariate}') in sset.free_symbols:
        warnings.warn('Covariate effect already exists')
        return model

    statistics = dict()
    statistics['mean'] = _calculate_mean(model.dataset, covariate)
    statistics['median'] = _calculate_median(model.dataset, covariate)
    statistics['std'] = _calculate_std(model.dataset, covariate)

    covariate_effect = _create_template(effect, model, covariate)
    thetas = _create_thetas(model, parameter, effect, covariate, covariate_effect.template)

    param_statement = sset.find_assignment(parameter)

    index = sset.index(param_statement)

    covariate_effect.apply(parameter, covariate, thetas, statistics)
    effect_statement = covariate_effect.create_effect_statement(operation, param_statement)

    statements = ModelStatements()

    statements += [s for s in covariate_effect.statistic_statements if s not in sset]
    statements.append(covariate_effect.template)
    statements.append(effect_statement)

    previous_effect = sset.find_assignment(parameter)
    cov_possible = [f'{parameter}{col_name}' for col_name in model.dataset.columns]

    if previous_effect.expression.args and all(
        arg.name in cov_possible for arg in previous_effect.expression.args if str(arg) != parameter
    ):
        effect_statement.expression = effect_statement.expression.subs(
            {parameter: previous_effect.expression}
        )
        sset.remove(previous_effect)

    for i, statement in enumerate(statements, 1):
        sset.insert(index + i, statement)

    model.statements = sset
    return model


def _create_thetas(model, parameter, effect, covariate, template):
    """Creates theta parameters and adds to parameter set of model.

    Number of parameters depends on how many thetas have been declared."""
    no_of_thetas = len(
        {str(sym) for sym in template.expression.free_symbols if str(sym).startswith('theta')}
    )

    pset = model.parameters

    theta_names = dict()

    if no_of_thetas == 1:
        inits = _choose_param_inits(effect, model.dataset, covariate)

        theta_name = f'POP_{parameter}{covariate}'
        pset.append(Parameter(theta_name, inits['init'], inits['lower'], inits['upper']))
        theta_names['theta'] = theta_name
    else:
        for i in range(1, no_of_thetas + 1):
            inits = _choose_param_inits(effect, model.dataset, covariate, i)

            theta_name = f'POP_{parameter}{covariate}_{i}'
            pset.append(Parameter(theta_name, inits['init'], inits['lower'], inits['upper']))
            theta_names[f'theta{i}'] = theta_name

    model.parameters = pset

    return theta_names


def _count_categorical(model, covariate):
    """Gets the number of individuals that has a level of categorical covariate."""
    idcol = model.dataset.pharmpy.id_label
    df = model.dataset.set_index(idcol)
    allcounts = df[covariate].groupby('ID').value_counts()
    allcounts.name = None  # To avoid collisions when resetting index
    counts = allcounts.reset_index().iloc[:, 1].value_counts()
    counts.sort_index(inplace=True)  # To make deterministic in case of multiple modes
    if model.dataset[covariate].isna().any():
        counts[np.nan] = 0
    return counts


def _calculate_mean(df, covariate, baselines=False):
    """Calculate mean. Can be set to use baselines, otherwise it is
    calculated first per individual, then for the group."""
    if baselines:
        return df[str(covariate)].mean()
    else:
        return df.groupby('ID')[str(covariate)].mean().mean()


def _calculate_median(df, covariate, baselines=False):
    """Calculate median. Can be set to use baselines, otherwise it is
    calculated first per individual, then for the group."""
    if baselines:
        return df.pharmpy.baselines[str(covariate)].median()
    else:
        return df.groupby('ID')[str(covariate)].median().median()


def _calculate_std(df, covariate, baselines=False):
    """Calculate median. Can be set to use baselines, otherwise it is
    calculated first per individual, then for the group."""
    if baselines:
        return df.pharmpy.baselines[str(covariate)].std()
    else:
        return df.groupby('ID')[str(covariate)].mean().std()


def _choose_param_inits(effect, df, covariate, index=None):
    """Chooses inits for parameters. If the effect is exponential, the
    bounds need to be dynamic."""
    init_default = 0.001

    inits = dict()

    cov_median = _calculate_median(df, covariate)
    cov_min = df[str(covariate)].min()
    cov_max = df[str(covariate)].max()

    lower, upper = _choose_bounds(effect, cov_median, cov_min, cov_max, index)

    if effect == 'exp':
        if lower > init_default or init_default > upper:
            init = (upper + lower) / 2
            if init == 0:
                init = upper / 5
        else:
            init = init_default
    elif effect == 'pow':
        init = init_default
    else:
        init = init_default

    inits['init'] = init
    inits['lower'] = lower
    inits['upper'] = upper

    return inits


def _choose_bounds(effect, cov_median, cov_min, cov_max, index=None):
    if effect == 'exp':
        min_diff = cov_min - cov_median
        max_diff = cov_max - cov_median

        lower_expected = 0.01
        upper_expected = 100

        if min_diff == 0 or max_diff == 0:
            return lower_expected, upper_expected
        else:
            log_base = 10
            lower = max(
                math.log(lower_expected, log_base) / max_diff,
                math.log(upper_expected, log_base) / min_diff,
            )
            upper = min(
                math.log(lower_expected, log_base) / min_diff,
                math.log(upper_expected, log_base) / max_diff,
            )
    elif effect == 'lin':
        if cov_median == cov_min:
            upper = 100000
        else:
            upper = 1 / (cov_median - cov_min)
        if cov_median == cov_max:
            lower = -100000
        else:
            lower = 1 / (cov_median - cov_max)
    elif effect == 'piece_lin':
        if cov_median == cov_min or cov_median == cov_max:
            raise Exception(
                'Median cannot be same as min or max, cannot use '
                'piecewise-linear parameterization.'
            )
        if index == 0:
            lower = -100000
            upper = 1 / (cov_median - cov_min)
        else:
            lower = 1 / (cov_median - cov_max)
            upper = 100000
    elif effect == 'pow':
        lower = -100
        upper = 100000
    else:
        lower = -100000
        upper = 100000
    return round(lower, 4), round(upper, 4)


def _create_template(effect, model, covariate):
    """Creates Covariate class objects with effect template."""
    if effect == 'lin':
        return CovariateEffect.linear()
    elif effect == 'cat':
        counts = _count_categorical(model, covariate)
        return CovariateEffect.categorical(counts)
    elif effect == 'piece_lin':
        return CovariateEffect.piecewise_linear()
    elif effect == 'exp':
        return CovariateEffect.exponential()
    elif effect == 'pow':
        return CovariateEffect.power()
    else:
        symbol = S('symbol')
        expression = sympy.sympify(effect)
        return CovariateEffect(Assignment(symbol, expression))


class CovariateEffect:
    """
    Covariate effect consisting of new assignments.

    Attributes
    ----------
    template
        Assignment based on covariate effect
    statistic_statements
        Dict with mean, median and standard deviation

    :meta private:

    """

    def __init__(self, template):
        self.template = template
        self.statistic_statements = []

    def apply(self, parameter, covariate, thetas, statistics):
        effect_name = f'{parameter}{covariate}'
        self.template.symbol = S(effect_name)

        self.template.subs(thetas)
        self.template.subs({'cov': covariate})

        template_str = [str(symbol) for symbol in self.template.free_symbols]

        if 'mean' in template_str:
            self.template.subs({'mean': f'{covariate}_MEAN'})
            s = Assignment(S(f'{covariate}_MEAN'), Float(statistics['mean'], 6))
            self.statistic_statements.append(s)
        if 'median' in template_str:
            self.template.subs({'median': f'{covariate}_MEDIAN'})
            s = Assignment(S(f'{covariate}_MEDIAN'), Float(statistics['median'], 6))
            self.statistic_statements.append(s)
        if 'std' in template_str:
            self.template.subs({'std': f'{covariate}_STD'})
            s = Assignment(S(f'{covariate}_STD'), Float(statistics['std'], 6))
            self.statistic_statements.append(s)

    def create_effect_statement(self, operation_str, statement_original):
        """Creates statement for addition or multiplication of covariate
        to parameter, e.g. (if parameter is CL and covariate is WGT):

            CL = CLWGT + TVCL*EXP(ETA(1))"""
        operation = self._get_operation(operation_str)

        symbol = statement_original.symbol
        expression = statement_original.symbol

        statement_new = Assignment(symbol, operation(expression, self.template.symbol))

        return statement_new

    @staticmethod
    def _get_operation(operation_str):
        """Gets sympy operation based on string"""
        if operation_str == '*':
            return mul
        elif operation_str == '+':
            return add

    @classmethod
    def linear(cls):
        """Linear continuous template (for continuous covariates)."""
        symbol = S('symbol')
        expression = 1 + S('theta') * (S('cov') - S('median'))
        template = Assignment(symbol, expression)

        return cls(template)

    @classmethod
    def categorical(cls, counts):
        """Linear categorical template (for categorical covariates)."""
        symbol = S('symbol')
        most_common = counts.idxmax()
        categories = list(counts.index)

        values = [1]
        conditions = [Eq(S('cov'), most_common)]

        for i, cat in enumerate(categories):
            if cat != most_common:
                if np.isnan(cat):
                    conditions += [Eq(S('cov'), S('NaN'))]
                    values += [1]
                else:
                    conditions += [Eq(S('cov'), cat)]
                    if len(categories) == 2:
                        values += [1 + S('theta')]
                    else:
                        values += [1 + S(f'theta{i}')]

        expression = Piecewise(*zip(values, conditions))

        template = Assignment(symbol, expression)

        return cls(template)

    @classmethod
    def piecewise_linear(cls):
        """Piecewise linear ("hockey-stick") template (for continuous
        covariates)."""
        symbol = S('symbol')
        values = [
            1 + S('theta1') * (S('cov') - S('median')),
            1 + S('theta2') * (S('cov') - S('median')),
        ]
        conditions = [Le(S('cov'), S('median')), Gt(S('cov'), S('median'))]
        expression = Piecewise((values[0], conditions[0]), (values[1], conditions[1]))

        template = Assignment(symbol, expression)

        return cls(template)

    @classmethod
    def exponential(cls):
        """Exponential template (for continuous covariates)."""
        symbol = S('symbol')
        expression = exp(S('theta') * (S('cov') - S('median')))
        template = Assignment(symbol, expression)

        return cls(template)

    @classmethod
    def power(cls):
        """Power template (for continuous covariates)."""
        symbol = S('symbol')
        expression = (S('cov') / S('median')) ** S('theta')
        template = Assignment(symbol, expression)

        return cls(template)

    def __str__(self):
        """String representation of class."""
        return str(self.template)
